#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functors library.

This is for running @Make instructions.

In the context of @Make we are thinking of each predicate as also a function
from the set of predicates to itself. The predicates that it uses in the
definition are arguments of this function.

Such functions that map predicates to predicates we call 'functors'.

At the moment any functor is a predicate, but in the future we may want to
introduce more abstact functors. E.g.: Apply(F, G) := F(G);
This could potentially be useful for limited recursion.
"""

import collections
import copy
import sys

if '.' not in __package__:
  from common import color
  from compiler.dialect_libraries import recursion_library
  from parser_py import parse
else:
  from ..common import color
  from ..compiler.dialect_libraries import recursion_library
  from ..parser_py import parse


class FunctorError(Exception):
  """Exception thrown when Make is bad."""

  def __init__(self, message, functor_name):
    super(FunctorError, self).__init__(message)
    self.functor_name = functor_name
    self.message = message

  def ShowMessage(self):
    print(color.Format('{underline}Making{end}:'), file=sys.stderr)
    print(self.functor_name, file=sys.stderr)
    print(color.Format('\n[ {error}Error{end} ] ') + self.message,
          file=sys.stderr)

def Walk(x, act):
  """Walking over a dictionary of lists, modifying and/or collecting info."""
  r = set()
  r |= set(act(x))
  if isinstance(x, list):
    for v in x:
      r |= Walk(v, act)
  if isinstance(x, dict):
    for k in x:
      r |= Walk(x[k], act)
  return r


class Functors(object):
  """A class for creating new functors."""

  def __init__(self, rules):
    self.rules = rules
    self.extended_rules = copy.deepcopy(rules)
    self.rules_of = parse.DefinedPredicatesRules(rules)
    self.predicates = set(self.rules_of)
    self.direct_args_of = self.BuildDirectArgsOf()
    self.args_of = {}
    self.creation_count = 0
    self.cached_calls = {}
    for p in self.predicates:
      self.ArgsOf(p)

  def CopyOfArgs(self):
    """Copying args is separated for profiling."""
    return {k: v for k, v in list(self.args_of.items())}

  def UpdateStructure(self, new_predicate):
    """Updates rules_of and args_of maps after extebded_rules update."""
    self.rules_of = parse.DefinedPredicatesRules(self.extended_rules)
    self.predicates = set(self.rules_of)
    self.direct_args_of = self.BuildDirectArgsOf()
    # Resetting args_of, to process the new predicates that were added.
    copied_args_of = self.CopyOfArgs()
    for predicate, args in copied_args_of.items():
      if new_predicate in args or new_predicate == predicate:
        del self.args_of[predicate]
    for p in self.predicates:
      self.ArgsOf(p)

  def ParseMakeInstruction(self, predicate, instruction):
    """Parses Make instruction from syntax tree."""
    error_message = 'Bad Make instruction: %s' % instruction
    if '1' not in instruction or '2' not in instruction:
      raise FunctorError(error_message, predicate)
    if 'predicate_name' not in instruction['1']:
      raise FunctorError(error_message, predicate)
    applicant = instruction['1']['predicate_name']
    args_map = {}
    for arg_name, arg_value_dict in instruction['2'].items():
      if 'predicate_name' not in arg_value_dict:
        raise FunctorError(error_message, predicate)
      args_map[arg_name] = arg_value_dict['predicate_name']
    return predicate, applicant, args_map

  def Describe(self):
    return 'DirectArgs: %s,\nArgs: %s' % (self.direct_args_of, self.args_of)

  def BuildDirectArgsOfWalk(self, x, act):
    """Factored for profiling."""
    return Walk(x, act)

  def BuildDirectArgsOf(self):
    """Builds a map of direct arguments of a functor."""
    def ExtractPredicateName(x):
      if isinstance(x, dict) and 'predicate_name' in x:
        return [x['predicate_name']]
      return []
    direct_args_of = {}
    for functor, rules in self.rules_of.items():
      args = set()
      for rule in rules:
        if 'body' in rule:
          args |= self.BuildDirectArgsOfWalk(rule['body'], ExtractPredicateName)
        args |= self.BuildDirectArgsOfWalk(rule['head']['record'],
                                           ExtractPredicateName)
      direct_args_of[functor] = args
    return direct_args_of

  def ArgsOf(self, functor):
    """Arguments of functor. Retrieving from cache, or computing."""
      
    if functor not in self.args_of:
      built_args = self.BuildArgs(functor)
      # Args could be incomplete due to recursive calls.
      # Checking for that.
      building_me = 'building_' + functor
      if building_me in built_args:
        built_args = built_args - {building_me}
      if any(a.startswith('building_') for a in built_args):
        return (a for a in built_args if not a.startswith('building_'))
      self.args_of[functor] = built_args

    return self.args_of[functor]

  def BuildArgs(self, functor):
    """Returning all arguments of a functor."""
    if functor not in self.direct_args_of:
      # Assuming this is built-in or table.
      return set()

    self.args_of[functor] = {'building_' + functor}

    result = set()
    queue = collections.deque(self.direct_args_of[functor])
    while queue:
      e = queue.popleft()
      result.add(e)
      for a in self.ArgsOf(e):
        if a not in result:
          queue.append(a)

    del self.args_of[functor]

    return result

  def AllRulesOf(self, functor):
    """Returning all rules relevant to a predicate."""
    result = []
    if functor not in self.rules_of:
      return result
    result.extend(self.rules_of[functor])
    for f in self.args_of[functor]:
      if f == functor:
        raise FunctorError('Failed to eliminate recursion of %s.' % functor,
                           functor)
      if f in self.rules_of:
        result.extend(self.rules_of[f])
    return copy.deepcopy(result)

  def Make(self, predicate, instruction):
    """Make a new predicate according to instruction."""
    self.CallFunctor(*self.ParseMakeInstruction(predicate, instruction))

  def MakeAll(self, predicate_to_instruction):
    """Making all required predicates."""
    # We needs to build them in the order of dependency.
    needs_building = set(
        self.ParseMakeInstruction(p, i)[0] for p, i in predicate_to_instruction)
    while needs_building:
      something_built = False
      for (new_predicate, instruction) in sorted(predicate_to_instruction):
        name, applicant, args_map = self.ParseMakeInstruction(new_predicate,
                                                              instruction)
        if (new_predicate not in needs_building or
            applicant in needs_building or
            (set(args_map.values()) & needs_building)):
          continue
        self.Make(new_predicate, instruction)
        something_built = True
        needs_building.remove(name)
      if needs_building and not something_built:
        raise FunctorError('Could not resolve Make order.',
                           str(needs_building))
    if False:
      with open('/tmp/functors.dot', 'w') as w:
        w.write('digraph Functors {\n')
        for f, args in self.direct_args_of.items():
          # For cleaner graph:
          if f.startswith('@') or f == '++?':
            continue
          for arg in args:
            if arg in ['||', '&&', '==', '<=', '>=', '<', '>', '!=', '++?',
                       '++', '+', '-', '*', '/', '%', '^', ' In ', '!',
                       'Agg+', 'Greatest']:
              continue  # For cleaner graph.
            w.write('"%s" -> "%s";\n' % (f, arg))
        w.write('}\n')

  def CollectAnnotations(self, predicates):
    """Collecting annotations of predictes."""
    predicates = set(predicates)
    result = []
    for annotation, rules in self.rules_of.items():
      if annotation not in ['@Limit', '@OrderBy', '@Ground',
                            '@NoInject']:
        continue
      for rule in rules:
        if rule['head']['record']['field_value'][0]['value']['expression'][
            'literal']['the_predicate']['predicate_name'] in predicates:
          result.append(rule)
    return copy.deepcopy(result)

  def CallKey(self, functor, args_map):
    """A string representing a call of a functor with arguments."""
    relevant_args = {k: v
                     for k, v in args_map.items()
                     if k in self.ArgsOf(functor)}
    args = ','.join('%s: %s' % (k, v) for k, v in sorted(relevant_args.items()))
    result = '%s(%s)' % (functor, args)
    return result

  def CallFunctor(self, name, applicant, args_map):
    """Calling a functor applicant(args_map), storing result to 'name'."""
    bad_args = set(args_map.keys()) - set(self.args_of[applicant])
    if bad_args:
      raise FunctorError(
          'Functor %s is applied to arguments %s, which it does not '
          'have.' % (applicant, color.Warn(','.join(bad_args))),
          name)
    self.creation_count += 1
    rules = self.AllRulesOf(applicant)
    args = set(args_map)
    rules = [r for r in rules
             if ((args & self.args_of[r['head']['predicate_name']]) or
                 r['head']['predicate_name'] == applicant)]
    if not rules:
      raise FunctorError(
          'Rules for %s when making %s are not found' % (applicant, name),
          name)
    # This eventually maps all args to substiturions, as well as all predictes
    # which use one of the args into a newly created predicate names.
    extended_args_map = copy.deepcopy(args_map)
    rules_to_update = []
    cache_update = {}
    predicates_to_annotate = set()
    for r in sorted(rules, key=str):
      rule_predicate_name = r['head']['predicate_name']
      if rule_predicate_name == applicant:
        extended_args_map[rule_predicate_name] = name
        rules_to_update.append(r)
        predicates_to_annotate.add(rule_predicate_name)
      else:
        # TODO: We should also remove predicates that become unused.
        if rule_predicate_name in args_map:
          continue
        call_key = self.CallKey(rule_predicate_name, args_map)
        if call_key in self.cached_calls:
          new_predicate_name = self.cached_calls[call_key]
          extended_args_map[rule_predicate_name] = new_predicate_name
        else:
          new_predicate_name = (
              rule_predicate_name + '_f%d' % self.creation_count)
          extended_args_map[rule_predicate_name] = new_predicate_name
          cache_update[call_key] = new_predicate_name
          rules_to_update.append(r)
          predicates_to_annotate.add(rule_predicate_name)

    rules = rules_to_update
    self.cached_calls.update(cache_update)
    # Collect annotations of all involved predicates.
    # Newly created predicates inherit annotations of predicates from which
    # they were created.
    # Arguments values of the functor do not inherit annotations of the
    # arguments, as that would collide with their performance in other
    # contexts.
    # In particular it means that functors should not use @Limit and @OrderBy
    # of the arguments to express the logic. Instead they should create simple
    # predicates reading from the agument and annotate those.
    annotations = self.CollectAnnotations(list(predicates_to_annotate))
    rules.extend(annotations)
    def ReplacePredicate(x):
      if isinstance(x, dict) and 'predicate_name' in x:
        if x['predicate_name'] in extended_args_map:
          x['predicate_name'] = extended_args_map[x['predicate_name']]
      return []
    Walk(rules, ReplacePredicate)
    self.extended_rules.extend(rules)
    self.UpdateStructure(name)

  def UnfoldRecursivePredicate(self, predicate, cover, depth, rules):   
    """Unfolds recurive predicate.""" 
    new_predicate_name = predicate + '_recursive'
    new_predicate_head_name = predicate + '_recursive_head'

    def ReplaceRecursivePredicate(x):
      if isinstance(x, dict) and 'predicate_name' in x:
        if x['predicate_name'] == predicate:
          x['predicate_name'] = new_predicate_name
      return []
    def ReplaceRecursiveHeadPredicate(x):
      if isinstance(x, dict) and 'predicate_name' in x:
        if x['predicate_name'] == predicate:
          x['predicate_name'] = new_predicate_head_name
      return []
    def ReplacerOfCoverMember(member):
      def Replace(x):
        if isinstance(x, dict) and 'predicate_name' in x:
          if x['predicate_name'] == member:
            x['predicate_name'] = member + '_recursive_head'
        return []
      return Replace

    for r in rules:
      if r['head']['predicate_name'] == predicate:
        r['head']['predicate_name'] = new_predicate_head_name
        Walk(r, ReplaceRecursivePredicate)
        for c in cover - {predicate}:
          Walk(r, ReplacerOfCoverMember(c))
      elif r['head']['predicate_name'] in cover:
        Walk(r, ReplaceRecursivePredicate)
        for c in cover - {predicate}:
          Walk(r, ReplacerOfCoverMember(c))
      elif (r['head']['predicate_name'][0] == '@' and
            r['head']['predicate_name'] != '@Make'):
        Walk(r, ReplaceRecursiveHeadPredicate)
        for c in cover - {predicate}:
          Walk(r, ReplacerOfCoverMember(c))        
      else:
        # This rule simply uses the predicate, keep the name.
        pass

    lib = recursion_library.GetRecursionFunctor(depth)
    lib = lib.replace('P', predicate)
    lib_rules = parse.ParseFile(lib)['rule']
    rules.extend(lib_rules)
    for c in cover - {predicate}:
      rename_lib = recursion_library.GetRenamingFunctor(c, predicate)
      rename_lib_rules = parse.ParseFile(rename_lib)['rule']
      rules.extend(rename_lib_rules)

  def UnfoldRecursions(self, depth_map):
    """Unfolds all recursions."""
    should_recurse, my_cover = self.RecursiveAnalysis(depth_map)
    new_rules = copy.deepcopy(self.rules)
    for p in should_recurse:
      depth = depth_map.get(p, {}).get('1', 8)
      self.UnfoldRecursivePredicate(p, my_cover[p], depth, new_rules)
    return new_rules

  def RecursiveAnalysis(self, depth_map):
    """Finds recursive cycles and predicates that would unfold them."""
    # TODO: Select unfolding predicates to guarantee unfolding.
    cover = []
    covered = set()
    deep = set(depth_map)
    for p, args in self.args_of.items():
      if p in args and p not in covered and '_MultBodyAggAux' not in p:
        c = {p}
        for p2 in args:
          if p in self.args_of[p2]:
            c.add(p2)
        cover.append(c)
        covered |= c

    my_cover = {}
    for c in cover:
      for p in c:
        my_cover[p] = c

    recursion_covered = set()
    should_recurse = []
    for c in cover:
      if c & deep:
        p = min(c & deep)
      else:
        p = min(c)
      should_recurse.append(p)
      recursion_covered |= my_cover[p]

    return should_recurse, my_cover


#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

library = """
->(left:, right:) = {arg: left, value: right};

Arrow(left, right) = arrow :-
  left == arrow.arg,
  right == arrow.value;

PrintToConsole(message) :- 1 == SqlExpr("PrintToConsole({message})", {message:});

ArgMin(arr) = Element(
    SqlExpr("ArgMin({a}, {v}, 1)", {a:, v:}), 0) :- Arrow(a, v) == arr;

ArgMax(arr) = Element(
    SqlExpr("ArgMax({a}, {v}, 1)", {a:, v:}), 0) :- Arrow(a, v) == arr;

ArgMinK(arr, k) = 
    SqlExpr("ArgMin({a}, {v}, {k})", {a:, v:, k:}) :-
  Arrow(a, v) == arr;

ArgMaxK(arr, k) =
    SqlExpr("ArgMax({a}, {v}, {k})", {a:, v:, k:}) :- Arrow(a, v) == arr;

ReadFile(filename) = SqlExpr("ReadFile({filename})", {filename:});

ReadJson(filename) = ReadFile(filename);

WriteFile(filename, content:) = SqlExpr("WriteFile({filename}, {content})",
                                        {filename:, content:});
"""



"""Provides connection to SQLite extended with UDFs needed by Logica."""

import csv
import io
import sys
import sqlite3
import heapq
import json

def DeFactoType(value):
  if isinstance(value, int) or isinstance(value, float):
    return 'number'
  else:
    return 'string'

class ArgMin:
  """ArgMin user defined aggregate function."""
  def __init__(self):
      self.result = []

  def step(self, arg, value, limit):
    if limit <= 0:
      raise Exception('ArgMin\'s limit must be positive.')
    if len(self.result) > 0:
      if DeFactoType(value) != DeFactoType(self.result[0][0]):
        raise Exception('ArgMin got incompatible values: %s vs %s' %
                        (repr(value), repr(self.result[0][0])))
    if len(self.result) < limit - 1:
      self.result.append((value, arg))
    elif len(self.result) == limit - 1:
      self.result.append((value, arg))
      heapq._heapify_max(self.result)
    elif len(self.result) == limit:
      if self.result[0][0] > value:
        heapq._heapreplace_max(self.result, (value, arg))
    else:
      print("ArgMin error:", self.result, arg, value, limit)
      raise Exception('ArgMin error')

  def finalize(self):
    return json.dumps([x[1] for x in sorted(self.result)])


class ArgMax:
  """ArgMax user defined aggregate functiom."""
  def __init__(self):
      self.result = []

  def step(self, arg, value, limit):
    if limit <= 0:
      raise Exception('ArgMax\'s limit must be positive.')
    if len(self.result) > 0:
      if DeFactoType(value) != DeFactoType(self.result[0][0]):
        raise Exception('ArgMax got incompatible values: %s vs %s' %
                        (repr(value), repr(self.result[0][0])))
    if len(self.result) < limit - 1:
      self.result.append((value, arg))
    elif len(self.result) == limit - 1:
      self.result.append((value, arg))
      heapq.heapify(self.result)
    elif len(self.result) == limit:
      if self.result[0][0] < value:
        heapq.heapreplace(self.result, (value, arg))
    else:
      print('ArgMax error:', self.result, arg, value, limit)
      raise Exception('ArgMax error')

  def finalize(self):
      return json.dumps([x[1] for x in reversed(sorted(self.result))])


def ArrayConcat(a, b):
  return json.dumps(json.loads(a) + json.loads(b))


def PrintToConsole(message):
  """User defined function printing to console."""
  print(message)
  return 1


def Join(array, separator):
  return separator.join(map(str, json.loads(array)))


def ReadFile(filename):
  try:
    with open(filename) as f:
      result = f.read()
  except:
    result = None
  return result


def WriteFile(filename, content):
  try:
    with open(filename, 'w') as w:
      w.write(content)
  except Exception as e:
    return str(e)
  return 'OK'


def ArtisticTable(header, rows):
  """ASCII art table for query output."""
  width = [0] * len(header)
  for r in [header] + rows:
    for i in range(len(r)):
      width[i] = max(width[i], len(str(r[i])))
  def Pad(s, w):
    return str(s) + ' ' * (w - len(str(s)))
  result = []
  top_line = '+-' + '-+-'.join('-' * w for w in width) + '-+'
  header_line = '| ' + ' | '.join(Pad(h, w) for h, w in zip(header, width)) + ' |'
  result = [top_line, header_line, top_line]
  for row in rows:
    result.append('| ' + ' | '.join(Pad(r, w) for r, w in zip(row, width)) + ' |')
  result.append(top_line)
  return '\n'.join(result)

def Csv(header, rows):
  """CSV query output."""
  stringio = io.StringIO()
  writer = csv.writer(stringio)
  writer.writerow(header)
  for row in rows:
    writer.writerow(row)
  return stringio.getvalue()

def SqliteConnect():
  con = sqlite3.connect(':memory:')
  con.create_aggregate('ArgMin', 3, ArgMin)
  con.create_aggregate('ArgMax', 3, ArgMax)
  con.create_function('PrintToConsole', 1, PrintToConsole)
  con.create_function('ARRAY_CONCAT', 2, ArrayConcat)
  con.create_function('JOIN_STRINGS', 2, Join)
  con.create_function('ReadFile', 1, ReadFile)
  con.create_function('WriteFile', 2, WriteFile)
  con.create_function('SQRT', 1, lambda x: float(x) ** 0.5)
  con.create_function('POW', 2, lambda x, p: float(x) ** p)
  sqlite3.enable_callback_tracebacks(True)
  return con


def RunSqlScript(statements, output_format):
  """Runs a sequence of statements, returning result of final."""
  assert statements, 'RunSqlScript requires non-empty statements list.'
  connect = SqliteConnect()
  cursor = connect.cursor()

  for s in statements[:-1]:
    cursor.executescript(s)
  cursor.execute(statements[-1])
  rows = cursor.fetchall()
  header = [d[0] for d in cursor.description]

  connect.close()
  if output_format == 'artistictable':
    result = ArtisticTable(header, rows)
  elif output_format == 'csv': 
    result = Csv(header, rows)
  else:
    assert False, 'Bad output format: %s' % output_format
  return result


def RunSQL(sql, output_format='artistictable'):
  """Running SQL with artistictable or csv output."""
  connect = SqliteConnect()
  cursor = connect.cursor()
  cursor.execute(sql)
  rows = cursor.fetchall()
  header = [d[0] for d in cursor.description]
  connect.close()
  if output_format == 'artistictable':
    result = ArtisticTable(header, rows)
  elif output_format == 'csv':
    result = Csv(header, rows)
  else:
    assert False, 'Bad output format: %s' % output_format
  return result

if __name__ == '__main__':
  c = SqliteConnect()
  print(RunSQL(sys.argv[1]))


#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A suite of tests for import functionality."""

from common import logica_test


def RunTest(name, src, golden, predicate, user_flags=None):
  """Run one test from this folder with TestManager."""
  logica_test.TestManager.RunTest(
      name,
      src="integration_tests/import_tests/" + src,
      golden="integration_tests/import_tests/" + golden,
      predicate=predicate,
      user_flags=user_flags)


def RunAll():
  """Run all the tests."""
  RunTest(
      name="canada_psql_test",
      src="canada_psql_test.l",
      golden="canada_psql_Consume.txt",
      predicate="TestConsume",
  )

  RunTest(
      name="functor_test",
      src="functor_test.l",
      golden="functor_test.txt",
      predicate="Test",
  )

  RunTest(
      name="canada_import_fraction_test",
      src="canada_test.l",
      golden="canada_ImportFraction.txt",
      predicate="TestImportFraction",
  )

  RunTest(
      name="canada_consume_test",
      src="canada_test.l",
      golden="canada_Consume.txt",
      predicate="TestConsume",
  )

  RunTest(
      name="canada_imported_engineer_test",
      src="canada_test.l",
      golden="canada_ImportedEngineer.txt",
      predicate="TestImportedEngineer",
  )




#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Compiler of a Logica expression to SQL."""

import copy

import json

if '.' not in __package__:
  from common import color
  from common.data import processed_functions
  from compiler import dialects
else:
  from ..common import color
  from ..common.data import processed_functions
  from ..compiler import dialects


class QL(object):
  """Class translating Logica expressions into SQL."""
  BUILT_IN_FUNCTIONS = {  # Please keep sorted alphabetically.
      # Casting.
      'ToFloat64': 'CAST(%s AS FLOAT64)',
      'ToInt64': 'CAST(%s AS INT64)',
      'ToUInt64': 'CAST(%s AS UINT64)',
      'ToString': 'CAST(%s AS STRING)',
      # Aggregation.
      'Aggr': '%s',  # Placeholder to use formulas for aggregation.
      'Agg+': 'SUM(%s)',
      'Agg++': 'ARRAY_CONCAT_AGG(%s)',
      'Container': '%s',
      'Count': 'APPROX_COUNT_DISTINCT(%s)',
      'ExactCount': 'COUNT(DISTINCT %s)',
      'List': 'ARRAY_AGG(%s)',
      'Median': 'APPROX_QUANTILES(%s, 2)[OFFSET(1)]',
      'SomeValue': 'ARRAY_AGG(%s IGNORE NULLS LIMIT 1)[OFFSET(0)]',
      # Other functions.
      '!': 'NOT %s',
      '-': '- %s',
      'Concat': 'ARRAY_CONCAT({0}, {1})',
      'Constraint': '%s',
      'DateAddDay': 'DATE_ADD({0}, INTERVAL {1} DAY)',
      'DateDiffDay': 'DATE_DIFF({0}, {1}, DAY)',
      'Element': '{0}[OFFSET({1})]',
      'Enumerate': ('ARRAY(SELECT STRUCT('
                    'ROW_NUMBER() OVER () AS n, x AS element) '
                    'FROM UNNEST(%s) as x)'),
      'IsNull': '(%s IS NULL)',
      'Join': 'ARRAY_TO_STRING(%s)',
      'Like': '({0} LIKE {1})',
      'Range': 'GENERATE_ARRAY(0, %s - 1)',
      'RangeOf': 'GENERATE_ARRAY(0, ARRAY_LENGTH(%s) - 1)',
      'Set': 'ARRAY_AGG(DISTINCT %s)',
      'Size': 'ARRAY_LENGTH(%s)',
      'Sort': 'ARRAY(SELECT x FROM UNNEST(%s) as x ORDER BY x)',
      'TimestampAddDays': 'TIMESTAMP_ADD({0}, INTERVAL {1} DAY)',
      'Unique': 'ARRAY(SELECT DISTINCT x FROM UNNEST(%s) as x ORDER BY x)',
      'ValueOfUnnested': '%s',
      # These functions are treated specially.
      'FlagValue': 'UNUSED',
      'Cast': 'UNUSED',
      'SqlExpr': 'UNUSED'
  }
  BUILT_IN_INFIX_OPERATORS = {
      '==': '%s = %s',
      '<=': '%s <= %s',
      '<': '%s < %s',
      '>=': '%s >= %s',
      '>': '%s > %s',
      '/': '(%s) / (%s)',
      '+': '(%s) + (%s)',
      '-': '(%s) - (%s)',
      '*': '(%s) * (%s)',
      '^': 'POW(%s, %s)',
      '!=': '%s != %s',
      '++': 'CONCAT(%s, %s)',
      'in': '%s IN UNNEST(%s)',
      '||': '%s OR %s',
      '&&': '%s AND %s',
      '%': 'MOD(%s, %s)'
  }
  BULK_FUNCTIONS = {}
  BULK_FUNCTIONS_ARITY_RANGE = {}

  # When adding any analytic functions please check that ConvertAnalytic
  # function handles them correctly.
  ANALYTIC_FUNCTIONS = {
      'CumulativeSum':
          'SUM({0}) OVER (PARTITION BY {1} ORDER BY {2} '
          'ROWS BETWEEN UNBOUNDED PRECEDING AND CURRENT ROW)',
      'CumulativeMax':
          'MAX({0}) OVER (PARTITION BY {1} ORDER BY {2} '
          'ROWS BETWEEN UNBOUNDED PRECEDING AND CURRENT ROW)',
      'CumulativeMin':
          'MIN({0}) OVER (PARTITION BY {1} ORDER BY {2} '
          'ROWS BETWEEN UNBOUNDED PRECEDING AND CURRENT ROW)',
      'WindowSum':
          'SUM({0}) OVER (PARTITION BY {1} ORDER BY {2} '
          'ROWS BETWEEN {3} PRECEDING AND CURRENT ROW)',
      'WindowMax':
          'MAX({0}) OVER (PARTITION BY {1} ORDER BY {2} '
          'ROWS BETWEEN {3} PRECEDING AND CURRENT ROW)',
      'WindowMin':
          'MIN({0}) OVER (PARTITION BY {1} ORDER BY {2} '
          'ROWS BETWEEN {3} PRECEDING AND CURRENT ROW)',
  }

  def __init__(self, vars_vocabulary, subquery_translator, exception_maker,
               flag_values, custom_udfs=None, dialect=None):
    """Initializes the instance.

    Args:
      vars_vocabulary: Dictionary mapping Logica variable to an SQL expression.
      subquery_translator: SubqueryTranslator to translate 'combine'
        expressions.
      exception_maker: Exception to raise if expression comilation error occurs.
      flag_values: Values of program flags.
      custom_udfs: A map from udf name to udf application string template.
      dialect: SQL dialect object.
    """
    self.dialect = dialect or dialects.BigQueryDialect()
    self.vocabulary = vars_vocabulary
    self.subquery_translator = subquery_translator
    # Using instance variables for test purposes.
    self.InstallBulkFunctionsOfStandardSQL()
    self.bulk_functions = self.BULK_FUNCTIONS
    self.bulk_function_arity_range = self.BULK_FUNCTIONS_ARITY_RANGE
    self.built_in_functions = copy.deepcopy(self.bulk_functions)
    self.built_in_functions.update(self.BUILT_IN_FUNCTIONS)
    self.built_in_functions.update(self.dialect.BuiltInFunctions())
    self.built_in_infix_operators = copy.deepcopy(
        self.BUILT_IN_INFIX_OPERATORS)
    self.built_in_infix_operators.update(self.dialect.InfixOperators())
    self.CleanOperatorsAndFunctions()
    self.exception_maker = exception_maker
    self.debug_undefined_variables = False
    # We set convert_to_json to convert arguments of annotations to Python
    # objects. This is hack-y. In the long run we plan to run them in
    # StandardSQL reference implementation and record the output.
    self.convert_to_json = False
    self.flag_values = flag_values
    self.custom_udfs = custom_udfs or {}

  def CleanOperatorsAndFunctions(self):
    def CleanDictionary(d):
      keys = list(d.keys())
      for k in keys:
        if d[k] is None:
          del d[k]
    for d in [self.built_in_infix_operators,
              self.built_in_functions]:
      CleanDictionary(self.built_in_infix_operators)

  @classmethod
  def BasisFunctions(cls):
    cls.InstallBulkFunctionsOfStandardSQL()
    return (
        set(cls.BUILT_IN_FUNCTIONS) |
        set(cls.BUILT_IN_INFIX_OPERATORS) |
        set(cls.BULK_FUNCTIONS) |
        set(cls.ANALYTIC_FUNCTIONS))

  @classmethod
  def InstallBulkFunctionsOfStandardSQL(cls):
    """Populates fields from processed_functions.csv."""
    if cls.BULK_FUNCTIONS:
      return
    def CamelCase(s):
      s = s.replace('.', '_')
      return ''.join(p[0].upper() + p[1:] for p in s.split('_'))
    reader = processed_functions.GetCsv()
    header = next(reader)

    for row in reader:
      row = dict(list(zip(header, row)))
      if row['function'][0] == '$':
        # TODO: Process operators.
        continue
      function_name = CamelCase(row['function'])
      cls.BULK_FUNCTIONS[function_name] = (
          '%s(%s)' % (row['sql_function'], '%s'))
      cls.BULK_FUNCTIONS_ARITY_RANGE[function_name] = (
          int(row['min_args']),
          float('inf')
          if row['has_repeated_args'] == '1' else int(row['max_args']))

  def BuiltInFunctionArityRange(self, f):
    """Returns arity of the built-in function."""
    assert f in self.built_in_functions
    if f in self.BUILT_IN_FUNCTIONS:
      if f == 'If':
        return (3, 3)
      arity_2_functions = ['RegexpExtract', 'Like',
                           'ParseTimestamp', 'FormatTimestamp',
                           'TimestampAddDays', 'Split', 'Element',
                           'Concat', 'DateAddDay', 'DateDiffDay',
                           'Join']
      if f in arity_2_functions:
        return (2, 2)
      return (1, 1)
    else:
      assert f in self.bulk_functions
      return self.bulk_function_arity_range[f]

  def If(self, args):
    assert len(args) == 3
    return 'IF(%s, %s, %s)' % tuple(args)

  def Function(self, f, args):
    args_list = [None] * len(args)
    for k, v in args.items():
      args_list[k] = str(v)
    if '%s' in f:
      return f % ', '.join(args_list)
    else:
      return f.format(*args_list)

  def Infix(self, op, args):
    return op % (args['left'], args['right'])

  def Subscript(self, record, subscript):
    if isinstance(subscript, int):
      subscript = 'col%d' % subscript
    return self.dialect.Subscript(record, subscript)

  def IntLiteral(self, literal):
    return str(literal['number'])

  def StrLiteral(self, literal):
    if self.dialect.Name() in ["PostgreSQL", "Presto", "Trino", "SqLite"]:
      # TODO: Do this safely.
      return '\'%s\'' % literal['the_string']
    return json.dumps(literal['the_string'], ensure_ascii=False)

  def ListLiteralInternals(self, literal):
    return ', '.join([self.ConvertToSql(e)
                      for e in literal['element']])

  def ListLiteral(self, literal):
    return self.dialect.ArrayPhrase() % self.ListLiteralInternals(literal)

  def BoolLiteral(self, literal):
    return literal['the_bool']

  def NullLiteral(self, literal):
    return literal['the_null']

  # Might be used for automatic program analysis.
  def PredicateLiteral(self, literal):
    if self.convert_to_json:
      return '{"predicate_name": "%s"}' % (literal['predicate_name'])
    return 'STRUCT("%s" AS predicate_name)' % literal['predicate_name']

  def Variable(self, variable):
    if variable['var_name'] in self.vocabulary:
      return self.vocabulary[variable['var_name']]
    else:
      if self.debug_undefined_variables:
        return 'UNDEFINED_%s' % variable['var_name']
      assert False, 'Found no interpretation for %s in %s' % (
          variable['var_name'], self.vocabulary)

  def ConvertRecord(self, args):
    result = {}
    for f_v in args['field_value']:
      assert 'expression' in f_v['value'], (
          'Bad record: %s' % args)
      result[f_v['field']] = self.ConvertToSql(f_v['value']['expression'])
    return result

  def RecordAsJson(self, record):
    json_field_values = []
    for f_v in record['field_value']:
      json_field_values.append('"{field}": {value}'.format(
          field=f_v['field'],
          value=self.ConvertToSql(f_v['value']['expression'])))
    return '{%s}' % ', '.join(json_field_values)

  def Record(self, record):
    if self.convert_to_json:
      return self.RecordAsJson(record)
    # TODO: Move this to dialects.py.
    if self.dialect.Name() == 'SqLite':
      arguments_str = ', '.join(
          "'%s', %s" % (f_v['field'],
                      self.ConvertToSql(f_v['value']['expression']) )
          for f_v in record['field_value'])
      return 'JSON_OBJECT(%s)' % arguments_str
    arguments_str = ', '.join(
        '%s AS %s' % (self.ConvertToSql(f_v['value']['expression']),
                      f_v['field'])
        for f_v in record['field_value'])
    if self.dialect.Name() == 'Trino':
      return '(SELECT %s)' % arguments_str
    return 'STRUCT(%s)' % arguments_str

  def GenericSqlExpression(self, record):
    """Converting SqlExpr to SQL."""
    top_record = self.ConvertRecord(record)
    if set(top_record) != set([0, 1]):
      raise self.exception_maker(
          'SqlExpr must have 2 positional arguments, got: %s' % top_record)
    if ('literal' not in record['field_value'][0]
        ['value']['expression'] or
        'the_string' not in
        record['field_value'][0]['value']['expression']['literal']):
      raise self.exception_maker(
          'SqlExpr must have first argument be string, got: %s' %
          top_record[0])

    template = (
        record['field_value'][0]['value']['expression']['literal']
        ['the_string']['the_string'])
    if 'record' not in record['field_value'][1]['value']['expression']:
      raise self.exception_maker(
          'Sectond argument of SqlExpr must be record literal. Got: %s' %
          top_record[1])
    args = self.ConvertRecord(
        record['field_value'][1]['value']['expression']['record'])
    return template.format(**args)

  def Implication(self, implication):
    when_then_clauses = []
    for cond_cons in implication['if_then']:
      when_then_clauses.append(
          'WHEN {cond} THEN {cons}'.format(
              cond=self.ConvertToSql(cond_cons['condition']),
              cons=self.ConvertToSql(cond_cons['consequence'])))
    when_then_clauses_str = ' '.join(when_then_clauses)
    otherwise = self.ConvertToSql(implication['otherwise'])
    return 'CASE %s ELSE %s END' % (when_then_clauses_str, otherwise)

  def ConvertAnalyticListArgument(self, expression):
    if ('literal' not in expression or
        'the_list' not in expression['literal']):
      raise self.exception_maker(
          'Analytic list argument must resolve to list literal, got: %s' %
          self.ConvertToSql(expression))
    return self.ListLiteralInternals(expression['literal']['the_list'])

  def ConvertAnalytic(self, call):
    """Converting analytic function call to SQL."""
    is_window = call['predicate_name'].startswith('Window')
    if len(call['record']['field_value']) != 3 + is_window:
      raise self.exception_maker(
          'Function %s must have %d arguments.' % (call['predicate_name'],
                                                   3 + is_window))
    aggregant = self.ConvertToSql(
        call['record']['field_value'][0]['value']['expression'])
    group_by = self.ConvertAnalyticListArgument(
        call['record']['field_value'][1]['value']['expression'])
    order_by = self.ConvertAnalyticListArgument(
        call['record']['field_value'][2]['value']['expression'])

    if is_window:
      window_size = self.ConvertToSql(
          call['record']['field_value'][3]['value']['expression'])

    if not is_window:
      return self.ANALYTIC_FUNCTIONS[call['predicate_name']].format(
          aggregant, group_by, order_by)
    else:
      return self.ANALYTIC_FUNCTIONS[call['predicate_name']].format(
          aggregant, group_by, order_by, window_size)

  def SubIfStruct(self, implication, subscript):
    """Optimizing SQL for subscript of an 'if' statement.
    
    Args:
      implication: Implication syntax tree.
      subscript: Subscript string.
      
    Returns:
      optimized SQL if all consequences are syntactic records, or None
      otherwise.
    """
    def GetValueOfField(field_values, field):
      for field_value in field_values:
        if field_value['field'] == field:
          return field_value['value']['expression']
      raise self.exception_maker(
          'Expected field %s missing in a record inside %s statement.' % (
          color.Warn(subscript), color.Warn('if')))
      assert False
    all_records = all(
        ('record' in if_then['consequence'])
        for if_then in implication['if_then'])
    if not (all_records and 'record' in implication['otherwise']):
      return None
    new_if_thens = []
    for if_then in implication['if_then']:
      new_if_then = copy.deepcopy(if_then)
      consequence = GetValueOfField(
          if_then['consequence']['record']['field_value'], subscript)
      new_if_then['consequence'] = consequence
      new_if_thens.append(new_if_then)
    new_otherwise = GetValueOfField(
        implication['otherwise']['record']['field_value'], subscript)
    new_expr = {
        'implication': {'if_then': new_if_thens, 'otherwise': new_otherwise}}
    return self.ConvertToSql(new_expr)

  def ConvertToSql(self, expression):
    """Converting Logica expression into SQL."""
    # print('EXPR:', expression)
    # Variables.
    if 'variable' in expression:
      return self.Variable(expression['variable'])

    # Literals.
    if 'literal' in expression:
      literal = expression['literal']
      if 'the_number' in literal:
        return self.IntLiteral(literal['the_number'])
      if 'the_string' in literal:
        return self.StrLiteral(literal['the_string'])
      if 'the_list' in literal:
        return self.ListLiteral(literal['the_list'])
      if 'the_bool' in literal:
        return self.BoolLiteral(literal['the_bool'])
      if 'the_null' in literal:
        return self.NullLiteral(literal['the_null'])
      if 'the_predicate' in literal:
        return self.PredicateLiteral(literal['the_predicate'])
      assert False, 'Logica bug: unsupported literal parsed: %s' % literal

    if 'call' in expression:
      call = expression['call']
      arguments = self.ConvertRecord(call['record'])
      if call['predicate_name'] in self.ANALYTIC_FUNCTIONS:
        return self.ConvertAnalytic(call)
      if call['predicate_name'] == 'SqlExpr':
        return self.GenericSqlExpression(call['record'])
      if call['predicate_name'] == 'Cast':
        if (len(arguments) != 2 or
            'literal' not in
            call['record']['field_value'][1]['value']['expression'] or
            'the_string' not in
            call['record']['field_value'][1]['value']['expression']['literal']):
          raise self.exception_maker(
              'Cast must have 2 arguments and the second argument must be a '
              'string literal.')
        cast_to = (
            call['record']['field_value'][1]['value']['expression']['literal']
            ['the_string']['the_string'])
        return 'CAST(%s AS %s)' % (
            self.ConvertToSql(
                call['record']['field_value'][0]['value']['expression']),
            cast_to)

      if call['predicate_name'] == 'FlagValue':
        if (len(arguments) != 1 or
            'literal' not in
            call['record']['field_value'][0]['value']['expression'] or
            'the_string' not in
            call['record']['field_value'][0]['value']['expression']['literal']):
          raise self.exception_maker(
              'FlagValue argument must be a string literal.')
        flag = (
            call['record']['field_value'][0]['value']['expression']['literal']
            ['the_string']['the_string'])
        if flag not in self.flag_values:
          raise self.exception_maker(
              'Unspecified flag: %s' % flag)
        return self.StrLiteral(
            {'the_string': self.flag_values[flag]})
      for ydg_f, sql_f in self.built_in_functions.items():
        if call['predicate_name'] == ydg_f:
          if not sql_f:
            raise self.exception_maker(
                'Function %s is not supported by %s dialect.' % (
                    color.Warn(ydg_f), color.Warn(self.dialect.Name())))
          if len(arguments) == 2 and ydg_f == '-':
            continue  # '-' is the only operator with variable arity.
          arity_range = self.BuiltInFunctionArityRange(ydg_f)
          if not arity_range[0] <= len(arguments) <= arity_range[1]:
            raise self.exception_maker(
                color.Format(
                    'Built-in function {warning}{ydg_f}{end} takes {a} '
                    'arguments, but {b} arguments were given.',
                    dict(ydg_f=ydg_f, a=arity_range,
                         b=len(arguments))))
          return self.Function(sql_f, arguments)

      for udf, udf_sql in self.custom_udfs.items():
        if call['predicate_name'] == udf:
          # TODO: Treatment of positional arguments should be
          # simplified everywhere.
          arguments = dict(
              (k, v) if isinstance(k, str) else ('col%d' % k, v)
              for k, v in arguments.items())
          try:
            result = udf_sql.format(**arguments)
          except KeyError:
            raise self.exception_maker(
                'Function %s call is inconsistent with its signature %s.' %
                (color.Warn(udf), udf_sql))
          return result

      for ydg_op, sql_op in self.built_in_infix_operators.items():
        if call['predicate_name'] == ydg_op:
          result = self.Infix(sql_op, arguments)
          result = '(' + result + ')'
          return result

    if 'subscript' in expression:
      sub = expression['subscript']
      subscript = sub['subscript']['literal']['the_symbol']['symbol']
      # TODO: Record literal and record of subscript should have
      # different keys.
      # Try to opimize and return the field from a record.
      if 'record' in sub['record']:
        for f_v in sub['record']['record']['field_value']:
          if f_v['field'] == subscript:
            # Optimizing and returning the field directly.
            return self.ConvertToSql(f_v['value']['expression'])
      # Trying to optmize subscript of implication.
      if 'implication' in sub['record']:
        simplified_sub = self.SubIfStruct(sub['record']['implication'],
                                          subscript)
        if simplified_sub:
          return simplified_sub
      # Couldn't optimize, just return the '.' expression.
      record = self.ConvertToSql(sub['record'])
      return self.Subscript(record, subscript)

    if 'record' in expression:
      record = expression['record']
      return self.Record(record)

    if 'combine' in expression:
      return '(%s)' % (
          self.subquery_translator.TranslateRule(expression['combine'],
                                                 self.vocabulary))

    if 'implication' in expression:
      implication = expression['implication']
      return self.Implication(implication)

    if 'call' in expression and 'predicate_name' in expression['call']:
      raise self.exception_maker(color.Format(
          'Unsupported supposedly built-in function: '
          '{warning}{predicate}{end}.', dict(
              predicate=expression['call']['predicate_name'])))
    assert False, (
        'Logica bug: expression %s failed to compile for unknown reason.' %
        str(expression))



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Running all integration tests."""

import sys
from common import logica_test
from integration_tests import run_tests as integration_tests
from integration_tests.import_tests import run_tests as import_tests


if 'golden_run' in sys.argv:
  logica_test.TestManager.SetGoldenRun(True)

if 'announce_tests' in sys.argv:
  logica_test.TestManager.SetAnnounceTests(True)

for a in sys.argv:
  if a.startswith('test_only='):
    logica_test.TestManager.SetRunOnlyTests(a.split('=')[1].split(','))

logica_test.PrintHeader()

integration_tests.RunAll()
import_tests.RunAll()



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding=utf-8
"""Compiler of predicates from Logica program to SQL."""

import collections
import copy
import re
import sys
import traceback
# Uncomment to debug.
# import pprint

import json

if '.' not in __package__:
  from common import color
  from compiler import dialects
  from compiler import expr_translate
  from compiler import functors
  from compiler import rule_translate
  from parser_py import parse
else:
  from ..common import color
  from ..compiler import dialects
  from ..compiler import expr_translate
  from ..compiler import functors
  from ..compiler import rule_translate
  from ..parser_py import parse

PredicateInfo = collections.namedtuple('PredicateInfo',
                                       ['embeddable'])
Ground = collections.namedtuple('Ground',
                                ['table_name', 'overwrite'])

xrange = range

def FormatSql(s): return s + ';'


class Logica(object):
  """Predicate execution accumulated data.

  * Data about DEFINE TABLE and EXPORT DATA accumulated over compilation.
  * Annotations of the universe.
  """

  def __init__(self):
    # TODO: Add a comment for each member.
    self.defines = []
    self.export_statements = []
    self.defines_and_exports = []
    self.table_to_defined_table_map = {}
    # Maps a @With'ed table to SQL implementing it.
    self.table_to_with_sql_map = {}
    # Maps a table T to a list of @With'ed tables for T's query.
    self.table_to_with_dependencies = collections.defaultdict(list)
    # Maps a grounded table T to a list of With'ed tables which were compiled
    # for T. With tables need to be re-compiled for each grounded tables so
    # that Ground dependencies of the With'ed tables are added.
    self.with_compilation_done_for_parent = collections.defaultdict(set)
    self.dependency_edges = []
    self.data_dependency_edges = []
    self.table_to_export_map = {}
    self.main_predicate_sql = None
    self.preamble = ''
    # Auxiliary structure for building dependency graph. At each moment of
    # execution this is an inverse path from the final predicate to the exported
    # predicate that is being built at the moment.
    self.workflow_predicates_stack = []
    self.flags_comment = ''
    self.compiling_udf = False
    self.annotations = None
    self.custom_udfs = None
    self.custom_udf_definitions = None
    self.main_predicate = None
    self.used_predicates = []
    self.dependencies_of = None

  def AddDefine(self, define):
    self.defines.append(define)

  def PredicateSpecificPreamble(self, predicate_name):
    needed_udfs = list(sorted([
        self.custom_udf_definitions[f]
        for f in self.dependencies_of[predicate_name]
        if f in self.custom_udf_definitions]))
    return '\n'.join(needed_udfs)

  def NeededUdfDefinitions(self):
    return list(sorted([self.custom_udf_definitions[f]
            for f in self.used_predicates if f in self.custom_udf_definitions]))

  def FullPreamble(self):
    return '\n'.join([self.flags_comment] + [self.preamble] + self.defines)

  def With(self, predicate_name):
    if self.compiling_udf:
      return False
    return self.annotations.With(predicate_name)


def Indent2(s):
  return '\n'.join('  ' + l for l in s.split('\n'))


def AnnotationError(message, annotation_value):
  raise rule_translate.RuleCompileException(
      message, annotation_value['__rule_text'])


class Annotations(object):
  """Utility to parse and retrieve predicate annotations."""
  ANNOTATING_PREDICATES = [
      '@Limit', '@OrderBy', '@Ground', '@Flag', '@DefineFlag',
      '@NoInject', '@Make', '@CompileAsTvf', '@With', '@NoWith',
      '@CompileAsUdf', '@ResetFlagValue', '@Dataset', '@AttachDatabase',
      '@Engine', '@Recursive'
  ]

  def __init__(self, rules, user_flags):
    # Extracting DefineFlags first, so that we can use flags in @Ground
    # annotations.
    self.annotations = self.ExtractAnnotations(
        rules, restrict_to=['@DefineFlag', '@ResetFlagValue'])
    self.user_flags = user_flags
    self.flag_values = self.BuildFlagValues()
    # [AnnotationName][PredicateName] -> List containing the annotation.
    self.annotations.update(
        **self.ExtractAnnotations(rules, flag_values=self.flag_values))
    self.CheckAnnotatedObjects(rules)

  def Preamble(self):
    """Query preamble based on annotations and flags."""
    preamble = ''
    attach_database_statements = self.AttachDatabaseStatements()
    if attach_database_statements:
      preamble += attach_database_statements + '\n\n'
    if self.Engine() == 'psql':
      preamble += (
          '-- Initializing PostgreSQL environment.\n'
          'set client_min_messages to warning;\n'
          'create schema if not exists logica_test;\n\n')
    return preamble

  def BuildFlagValues(self):
    """Building values by overriding defaults with user flags."""
    default_values = {}
    for flag, a in self.annotations['@DefineFlag'].items():
      default_values[flag] = a.get('1', '${%s}' % flag)
    programmatic_flag_values = {}
    for flag, a in self.annotations['@ResetFlagValue'].items():
      programmatic_flag_values[flag] = a.get('1', '${%s}' % flag)

    if not set(self.user_flags) <= set(default_values):
      raise rule_translate.RuleCompileException(
          'Undefined flags used: %s' % list(
              set(self.user_flags) - set(default_values)),
          str(set(self.user_flags) - set(default_values)))
    flag_values = default_values
    flag_values.update(**programmatic_flag_values)
    flag_values.update(**self.user_flags)
    return flag_values

  def NoInject(self, predicate_name):
    return predicate_name in self.annotations['@NoInject']

  def OkInjection(self, predicate_name):
    # Current annotations make predicate non-injectible. This doesn't have
    # to be the case for an arbitrary annotation.
    if (self.OrderBy(predicate_name) or self.LimitOf(predicate_name) or
        self.Ground(predicate_name) or self.NoInject(predicate_name) or
        self.ForceWith(predicate_name)):
      return False
    return True

  def AttachedDatabases(self):
    result = {}
    for k, v in self.annotations['@AttachDatabase'].items():
      if '1' not in v:
        print('>>', v)
        AnnotationError('@AttachDatabase must have a single argument.',
                        v)
      result[k] = v['1']
    if (self.Engine() == 'sqlite'
        and 'logica_test' not in result
        and '@Ground' in self.annotations and
        self.annotations['@Ground']):
      result['logica_test'] = ':memory:'
    return result

  def AttachDatabaseStatements(self):
    return '\n'.join(
        'ATTACH DATABASE \'%s\' AS %s;' % (v, k)
        for k, v in self.AttachedDatabases().items())

  def CompileAsUdf(self, predicate_name):
    result = predicate_name in self.annotations['@CompileAsUdf']
    if result and self.TvfSignature(predicate_name):
      raise rule_translate.RuleCompileException(
          'A predicate can not be UDF and TVF at the '
          'same time %s.' % predicate_name,
          'Predicate: ' + predicate_name)
    return result

  def TvfSignature(self, predicate_name):
    """Table valued function signature of a predicate."""
    if predicate_name not in self.annotations['@CompileAsTvf']:
      return None
    annotation = self.annotations['@CompileAsTvf'][predicate_name]['1']
    arguments = [x['predicate_name'] for x in annotation]
    signature = ', '.join('%s ANY TABLE' % a for a in arguments)
    return 'CREATE TEMP TABLE FUNCTION %s(%s) AS ' % (predicate_name,
                                                      signature)

  def LimitOf(self, predicate_name):
    """Limit of the query corresponding to the predicate as per annotation."""
    if predicate_name not in self.annotations['@Limit']:
      return None
    annotation = FieldValuesAsList(self.annotations['@Limit'][predicate_name])
    if (len(annotation) != 1 or
        not isinstance(annotation[0], int)):
      raise rule_translate.RuleCompileException(
          'Bad limit specification for predicate %s.' % predicate_name,
          'Predicate: ' + predicate_name)
    return annotation[0]

  def OrderBy(self, predicate_name):
    if predicate_name not in self.annotations['@OrderBy']:
      return None
    return FieldValuesAsList(self.annotations['@OrderBy'][predicate_name])

  def Dataset(self):
    return self.ExtractSingleton('@Dataset', 'logica_test')

  def Engine(self):
    engine = self.ExtractSingleton('@Engine', 'bigquery')
    if engine not in dialects.DIALECTS:
      AnnotationError('Unrecognized engine: %s' % engine,
                      self.annotations['@Engine'][engine])
    return engine

  def ExtractSingleton(self, annotation_name, default_value):
    if not self.annotations[annotation_name]:
      return default_value
    results = list(self.annotations[annotation_name].keys())
    if len(results) > 1:
      raise rule_translate.RuleCompileException(
          'Single %s must be provided. Provided: %s' % (
              annotation_name, results),
          self.annotations[annotation_name][results[0]]['__rule_text'])
    return results[0]

  def Ground(self, predicate_name):
    """Returns Ground (physical file) associated with the predicate."""
    if predicate_name not in self.annotations['@Ground']:
      return None
    annotation = self.annotations['@Ground'][predicate_name]
    table_name = annotation.get('1', self.Dataset() + '.' + predicate_name)
    overwrite = annotation.get('overwrite', True)
    return Ground(table_name=table_name, overwrite=overwrite)

  def ForceWith(self, predicate_name):
    """Return true if the predicate has been explicitly marked @With."""
    return predicate_name in self.annotations['@With']

  def ForceNoWith(self, predicate_name):
    """Return true if the predicate has been explicitly marked @NoWith."""
    return predicate_name in self.annotations['@NoWith']

  def With(self, predicate_name):
    """Return whether this predicate should be compiled to a WITH-table.

    This only applies if the predicate is not inlined earlier in the flow.
    """
    is_with = self.ForceWith(predicate_name)
    is_nowith = self.ForceNoWith(predicate_name)
    if is_with and is_nowith:
      raise rule_translate.RuleCompileException(
          color.Format('Predicate is annotated both with @With and @NoWith.'),
          'Predicate: %s' % predicate_name)
    if is_with:
      return True
    if is_nowith or self.Ground(predicate_name):
      return False
    # TODO: return false for predicates that will be injected.
    return True

  def LimitClause(self, predicate_name):
    limit = self.LimitOf(predicate_name)
    if limit:
      return ' LIMIT %d' % limit
    else:
      return ''

  def OrderByClause(self, predicate_name):
    """Returns ORDER BY clause."""
    order_by = self.OrderBy(predicate_name)
    result = []
    if order_by:
      for i in range(len(order_by) - 1):
        if order_by[i+1] != 'DESC':
          result.append(order_by[i] + ',')
        else:
          result.append(order_by[i])
      result.append(order_by[-1])
      return ' ORDER BY ' + ' '.join(result)
    else:
      return ''

  def CheckAnnotatedObjects(self, rules):
    """Verify that annotations are applied to existing predicates."""
    all_predicates = {rule['head']['predicate_name'] for rule in rules} | set(
        self.annotations['@Ground']) | set(self.annotations['@Make'])
    for annotation_name in self.annotations:
      if annotation_name in {'@Limit', '@OrderBy',
                             '@NoInject', '@CompileAsTvf', '@With', '@NoWith',
                             '@CompileAsUdf'}:
        for annotated_predicate in self.annotations[annotation_name]:
          if annotated_predicate not in all_predicates:
            rule_text = self.annotations[annotation_name][annotated_predicate][
                '__rule_text']
            RaiseCompilerError(
                'Annotation %s must be applied to an existing predicate, but '
                'it was applied to a non-existing predicate %s.' %
                (color.Warn(annotation_name), color.Warn(annotated_predicate)),
                rule_text)

  @classmethod
  def ExtractAnnotations(cls, rules, restrict_to=None, flag_values=None):
    """Extracting annotations from the rules."""
    result = {p: collections.OrderedDict() for p in cls.ANNOTATING_PREDICATES}
    for rule in rules:
      rule_predicate = rule['head']['predicate_name']
      if restrict_to and rule_predicate not in restrict_to:
        continue
      if (rule_predicate[0] == '@' and
          rule_predicate not in cls.ANNOTATING_PREDICATES):
        raise rule_translate.RuleCompileException(
            'Only {0} and {1} special predicates are allowed.'.format(
                ', '.join(cls.ANNOTATING_PREDICATES[:-1]),
                cls.ANNOTATING_PREDICATES[-1]),
            rule['full_text'])
      if rule_predicate in cls.ANNOTATING_PREDICATES:
        rule_text = rule['full_text']
        # pylint: disable=cell-var-from-loop
        def ThrowException(*args, **xargs):
          _ = args
          _ = xargs
          if rule_predicate == '@Make':
            # pylint: disable=raising-format-tuple
            raise rule_translate.RuleCompileException(
                'Incorrect syntax for functor call. '
                'Functor call to be made as\n'
                '  R := F(A: V, ...)\n'
                'or\n'
                '  @Make(R, F, {A: V, ...})\n'
                'Where R, F, A\'s and V\'s are all '
                'predicate names.',
                rule_text)
          else:
            raise rule_translate.RuleCompileException(
                'Can not understand annotation.',
                rule_text)

        class Thrower(object):

          def __contains__(self, key):
            if rule_predicate == '@Make':
              ThrowException()
              return
            raise rule_translate.RuleCompileException(
                'Annotation may not use variables, but this one uses '
                'variable %s.' % (
                    color.Warn(key)),
                rule_text)
        flag_values = flag_values or Thrower()
        ql = expr_translate.QL(Thrower(), ThrowException, ThrowException,
                               flag_values)
        ql.convert_to_json = True

        annotation = rule['head']['predicate_name']
        field_values_json_str = ql.ConvertToSql(
            {'record': rule['head']['record']})
        try:
          field_values = json.loads(field_values_json_str)
        except:
          raise rule_translate.RuleCompileException(
              'Could not understand arguments of annotation.',
              rule['full_text'])
        if ('0' in field_values and
            isinstance(field_values['0'], dict) and
            'predicate_name' in field_values['0']):
          subject = field_values['0']['predicate_name']
        else:
          subject = field_values['0']
        del field_values['0']
        if rule_predicate in ['@OrderBy', '@Limit', '@NoInject']:
          field_values_list = FieldValuesAsList(field_values)
          if field_values_list is None:
            raise rule_translate.RuleCompileException(
                '@OrderBy and @Limit may only have positional '
                'arguments.', rule['full_text'])
          if rule_predicate == '@Limit' and len(field_values_list) != 1:
            raise rule_translate.RuleCompileException(
                'Annotation @Limit must have exactly two arguments: '
                'predicate and limit.', rule['full_text'])
        updated_annotation = result.get(annotation, {})
        field_values['__rule_text'] = rule['full_text']
        if subject in updated_annotation:
          raise rule_translate.RuleCompileException(
              color.Format(
                  '{annotation} annotates {warning}{subject}{end} more '
                  'than once: {before}, {after}',
                  dict(annotation=annotation,
                       subject=subject,
                       before=updated_annotation[subject]['__rule_text'],
                       after=field_values['__rule_text'])), rule['full_text'])
        updated_annotation[subject] = field_values
        result[annotation] = updated_annotation
    return result


class LogicaProgram(object):
  """Representing a Logica program.

  Can produce SQL for predicates.
  """

  def __init__(self, rules, table_aliases=None, user_flags=None):
    """Initializes the program.

    Args:
      rules: A list of dictionary representations of parsed Logica rules.
      table_aliases: A map from an undefined Logica predicate name to a
        BigQuery table name. This table will be used in place of predicate.
      user_flags: Dictionary of user specified flags.
    """
    rules = self.UnfoldRecursion(rules)

    # TODO: Should allocator be a member of Logica?
    self.preparsed_rules = rules
    self.rules = []
    self.defined_predicates = set()
    self.dollar_params = list(self.ExtractDollarParams(rules))
    self.table_aliases = table_aliases or {}
    self.execution = None
    self.user_flags = user_flags or {}
    self.annotations = Annotations(rules, self.user_flags)
    self.flag_values = self.annotations.flag_values
    # Dictionary custom_udfs maps function name to a format string to use
    # in queries.
    self.custom_udfs = collections.OrderedDict()
    # Dictionary custom_udf_definitions maps function name to SQL defining the
    # function.
    self.custom_udf_definitions = collections.OrderedDict()
    if not set(self.dollar_params) <= set(self.flag_values):
      raise rule_translate.RuleCompileException(
          'Parameters %s are undefined.' % (
              list(set(self.dollar_params) - set(self.flag_values))),
          str(list(set(self.dollar_params) - set(self.flag_values))))
    self.functors = None

    # Extending rules with functors.
    extended_rules = self.RunMakes(rules)  # Populates self.functors.

    # Extending rules with the library of the dialect.
    library_rules = parse.ParseFile(
        dialects.Get(self.annotations.Engine()).LibraryProgram())['rule']
    extended_rules.extend(library_rules)

    for rule in extended_rules:
      predicate_name = rule['head']['predicate_name']
      self.defined_predicates.add(predicate_name)
      self.rules.append((predicate_name, rule))
    # We need to recompute annotations, because 'Make' created more rules and
    # annotations.
    self.annotations = Annotations(extended_rules, self.user_flags)
    # Build udfs, populating custom_udfs and custom_udf_definitions.
    self.BuildUdfs()
    # Function compilation may have added irrelevant defines:
    self.execution = None

    if False:
      self.RunTypechecker()

  def UnfoldRecursion(self, rules):
    annotations = Annotations(rules, {})
    f = functors.Functors(rules)
    return f.UnfoldRecursions(annotations.annotations.get('@Recursive', {}))

  def BuildUdfs(self):
    """Build UDF definitions."""
    self.InitializeExecution('@FunctionsCheck')
    self.execution.compiling_udf = True
    remove_udfs = False
    for f in self.annotations.annotations['@CompileAsUdf']:
      if not remove_udfs:
        self.custom_udfs[f] = 'DUMMY()'  # For dummy function definitions.
    # First time we create proper function definitions, when we use them.
    # TODO: If we extract signature building then we don't have to
    # compile twice.
    for _ in range(2):
      for f in self.annotations.annotations['@CompileAsUdf']:
        application, sql = self.FunctionSql(f, internal_mode=True)
        if not remove_udfs:
          self.custom_udfs[f] = application
          self.custom_udf_definitions[f] = sql

  def NewNamesAllocator(self):
    return rule_translate.NamesAllocator(custom_udfs=self.custom_udfs)

  def RunTypechecker(self):
    """Checks the program for type-correctness.

    Raises:
      TypeInferenceError if there are any type errors.
    """
    inference.CheckTypes(self.rules)

  def RunMakes(self, rules):
    """Runs @Make instructions."""
    if '@Make' not in self.annotations.annotations:
      return rules
    self.functors = functors.Functors(rules)
    self.functors.MakeAll(list(self.annotations.annotations['@Make'].items()))
    return self.functors.extended_rules

  @classmethod
  def ExtractDollarParamsFromString(cls, s):
    params = re.findall(r'[$][{](.*?)[}]', s)
    return set(p for p in params  # Built-in, we shouldn't override them.
               if not (p.startswith('YYYY') or p == 'MM' or p == 'DD'))

  def ExtractDollarParams(self, r):
    """Returns a set of dollar params, like ${date}."""
    # TODO: Refactor the tree scanning into a helper class.
    if isinstance(r, dict):
      member_index = sorted(r.keys())
    elif isinstance(r, list):
      member_index = range(len(r))
    else:
      if isinstance(r, str):
        return self.ExtractDollarParamsFromString(r)
      return set()

    result = set()
    for k in member_index:
      result |= self.ExtractDollarParams(r[k])
    return result

  def __str__(self):
    return str(self.preparsed_rules)

  def GetPredicateRules(self, predicate_name):
    for (n, r) in self.rules:
      if n == predicate_name:
        yield r

  def PredicateSql(self, name, allocator=None, external_vocabulary=None):
    """Producing SQL for a predicate."""
    # Load proto if necessary.
    rules = list(self.GetPredicateRules(name))
    if len(rules) == 1:
      [rule] = rules
      result = (
          self.SingleRuleSql(rule, allocator, external_vocabulary) +
          self.annotations.OrderByClause(name) +
          self.annotations.LimitClause(name))
      if result.startswith('/* nil */'):
        raise rule_translate.RuleCompileException(
          'Single rule is nil for predicate %s. '
          'Recursion unfolding failed.' % color.Warn(name),
          rule['full_text'])
      return result
    elif len(rules) > 1:
      rules_sql = []
      for rule in rules:
        if 'distinct_denoted' in rule:
          raise rule_translate.RuleCompileException(
              color.Format(
                  'For distinct denoted predicates multiple rules are not '
                  'currently supported. Consider taking '
                  '{warning}union of bodies manually{end}, if that was what '
                  'you intended.'), rule['full_text'])
        single_rule_sql = self.SingleRuleSql(
            rule, allocator, external_vocabulary)
        if not single_rule_sql.startswith('/* nil */'):
          rules_sql.append('\n%s\n' %
                          Indent2(single_rule_sql))
      if not rules_sql:
        raise rule_translate.RuleCompileException(
          'All disjuncts are nil for predicate %s.' % color.Warn(name),
          rule['full_text'])

      rules_sql = ['\n'.join('  ' + l for l in r.split('\n'))
                   for r in rules_sql]
      return 'SELECT * FROM (\n%s\n) AS UNUSED_TABLE_NAME %s %s' % (
          ' UNION ALL\n'.join(rules_sql),
          self.annotations.OrderByClause(name),
          self.annotations.LimitClause(name))
    else:
      raise rule_translate.RuleCompileException(
          color.Format(
              'No rules are defining {warning}{name}{end}, but compilation '
              'was requested.', dict(name=name)),
          r'        ¯\_(ツ)_/¯')

  @classmethod
  def TurnPositionalIntoNamed(self, select):
    """Auxiliary method transforming RuleStructure select dict.

    Args:
      select: RuleStructure select dictionary, i.e. a map from variable name
        to Logica expression of a value dictionary.

    Returns:
      newly generated select dict, where positional arguments where replaced
      with the names of variables that were standing there.
    """
    new_select = collections.OrderedDict()
    # Make UDF use positional arguments as well.
    for v in select:
      if isinstance(v, int):
        # Check for error.
        new_select[select[v]['variable']['var_name']] = select[v]
      else:
        new_select[v] = select[v]
    return new_select

  def FunctionSql(self, name, allocator=None, internal_mode=False):
    """Print formatted SQL function creation statement."""
    # TODO: Refactor this into FunctionSqlInternal and FunctionSql.
    if not allocator:
      allocator = self.NewNamesAllocator()

    rules = list(self.GetPredicateRules(name))

    # Check that the predicate is defined via a single rule.
    if not rules:
      raise rule_translate.RuleCompileException(
          color.Format(
              'No rules are defining {warning}{name}{end}, but compilation '
              'was requested.', dict(name=name)),
          r'        ¯\_(ツ)_/¯')
    elif len(rules) > 1:
      raise rule_translate.RuleCompileException(
          color.Format(
              'Predicate {warning}{name}{end} is defined by more than 1 rule '
              'and can not be compiled into a function.', dict(name=name)),
          '\n\n'.join(r['full_text'] for r in rules))
    [rule] = rules

    # Extract structure and assert that it is isomorphic to a function.
    s = rule_translate.ExtractRuleStructure(rule,
                                            external_vocabulary=None,
                                            names_allocator=allocator)

    udf_variables = [v if isinstance(v, str) else 'col%d' % v
                     for v in s.select if v != 'logica_value']
    s.select = self.TurnPositionalIntoNamed(s.select)

    variables = [v for v in s.select if v != 'logica_value']
    if 0 in variables:
      raise rule_translate.RuleCompileException(
          color.Format(
              'Predicate {warning}{name}{end} must have all aruments named for '
              'compilation as a function.',
              dict(name=name)
          ), rule['full_text'])
    for v in variables:
      if ('variable' not in s.select[v] or
          s.select[v]['variable']['var_name'] != v):
        raise rule_translate.RuleCompileException(
            color.Format(
                'Predicate {warning}{name}{end} must not rename arguments '
                'for compilation as a function.',
                dict(name=name)
            ), rule['full_text'])

    vocabulary = {v: v for v in variables}
    s.external_vocabulary = vocabulary
    self.RunInjections(s, allocator)
    s.ElliminateInternalVariables(assert_full_ellimination=True)
    s.UnificationsToConstraints()
    sql = s.AsSql(subquery_encoder=self.MakeSubqueryTranslator(allocator))
    if s.constraints or s.unnestings or s.tables:
      raise rule_translate.RuleCompileException(
          color.Format(
              'Predicate {warning}{name}{end} is not a simple function, but '
              'compilation as function was requested. Full SQL:\n{sql}',
              dict(name=name, sql=sql)
          ), rule['full_text'])
    if 'logica_value' not in s.select:
      raise rule_translate.RuleCompileException(
          color.Format(
              'Predicate {warning}{name}{end} does not have a value, but '
              'compilation as function was requested. Full SQL:\n%s' %
              sql
          ), rule['full_text'])

    # pylint: disable=g-long-lambda
    # Compile the function!
    ql = expr_translate.QL(vocabulary,
                           self.MakeSubqueryTranslator(allocator),
                           lambda message:
                           rule_translate.RuleCompileException(
                               message, rule['full_text']),
                           self.flag_values,
                           custom_udfs=self.custom_udfs,
                           dialect=self.execution.dialect)
    value_sql = ql.ConvertToSql(s.select['logica_value'])

    sql = 'CREATE TEMP FUNCTION {name}({signature}) AS ({value})'.format(
        name=name,
        signature=', '.join('%s ANY TYPE' % v for v in variables),
        value=value_sql)

    sql = FormatSql(sql)

    if internal_mode:
      return ('%s(%s)' % (name, ', '.join('{%s}' % v for v in udf_variables)),
              sql)

    return sql

  def InitializeExecution(self, main_predicate):
    """Initialize self.execution."""
    self.execution = Logica()
    self.execution.workflow_predicates_stack.append(main_predicate)
    self.execution.preamble = self.annotations.Preamble()
    self.execution.annotations = self.annotations
    self.execution.custom_udfs = self.custom_udfs
    self.execution.custom_udf_definitions = self.custom_udf_definitions
    self.execution.main_predicate = main_predicate
    self.execution.used_predicates = self.functors.args_of.get(main_predicate,
                                                               [])
    self.execution.dependencies_of = self.functors.args_of
    self.execution.dialect = dialects.Get(self.annotations.Engine())

  def FormattedPredicateSql(self, name, allocator=None):
    """Printing top-level formatted SQL statement with defines and exports."""
    self.InitializeExecution(name)
    if self.flag_values and False:  # TODO: Control flag printing.
      flags_str_lines = ['# Logica flags:']
      for flag, value in sorted(self.flag_values.items()):
        flags_str_lines.append('#   %s = %s' % (flag, value.encode('utf-8')))
      self.execution.flags_comment = '\n'.join(flags_str_lines) + '\n\n'

    if self.annotations.CompileAsUdf(name):
      self.execution.compiling_udf = True
      sql = self.FunctionSql(name, allocator)
    else:
      sql = self.PredicateSql(name, allocator)

    assert self.execution.workflow_predicates_stack == [name], (
        'Logica internal error: unexpected workflow stack: %s' %
        self.execution.workflow_predicates_stack)

    # Wrap query in with
    with_signature = self.GenerateWithClauses(name)
    if with_signature:
      sql = '{}\n{}'.format(with_signature, sql)
    self.execution.table_to_export_map[name] = sql
    defines_and_exports = self.execution.preamble
    udf_definitions = self.execution.NeededUdfDefinitions()
    if udf_definitions:
      defines_and_exports += '\n\n'.join(udf_definitions)
      defines_and_exports += '\n\n'

    if self.execution.defines_and_exports:
      defines_and_exports += '\n\n'.join(self.execution.defines_and_exports)
      defines_and_exports += '\n\n'

    sql = self.UseFlagsAsParameters(sql)  # To avoid formatting errors.

    # Append TVF signature.
    tvf_signature = self.annotations.TvfSignature(name)
    if tvf_signature:
      sql = tvf_signature + '\n' + sql

    self.execution.main_predicate_sql = sql
    formatted_sql = (
        self.execution.flags_comment +
        defines_and_exports +
        FormatSql(sql))
    if True:
      for k, v in self.execution.table_to_export_map.items():
        self.execution.table_to_export_map[k] = self.UseFlagsAsParameters(v)
      for i, d in enumerate(self.execution.defines):
        self.execution.defines[i] = self.UseFlagsAsParameters(d)
      self.execution.flags_comment = self.UseFlagsAsParameters(
          self.execution.flags_comment)
      self.execution.main_predicate_sql = self.UseFlagsAsParameters(
        self.execution.main_predicate_sql)
      return self.UseFlagsAsParameters(formatted_sql)
    else:
      return formatted_sql

  def UseFlagsAsParameters(self, sql):
    """Running flag substitution in a loop to the fixed point."""
    # We do it in a loop to deal with flags that refer to other flags.
    prev_sql = ''
    num_subs = 0
    while sql != prev_sql:
      num_subs += 1
      prev_sql = sql
      if num_subs > 100:
        raise rule_translate.RuleCompileException(
            'You seem to have recursive flags. It is disallowed.',
            'Flags:\n' +
            '\n'.join('--{0}={1}'.format(*i)
                      for i in self.flag_values.items()))
      # Do the substitution!
      for flag, value in self.flag_values.items():
        sql = sql.replace('${%s}' % flag, value)
    return sql

  def RunInjections(self, s, allocator):
    iterations = 0
    while True:
      iterations += 1
      if iterations > sys.getrecursionlimit():
        raise rule_translate.RuleCompileException(
            RecursionError(),
            s.full_rule_text)

      new_tables = collections.OrderedDict()
      for table_name_rsql, table_predicate_rsql in s.tables.items():
        rules = list(self.GetPredicateRules(table_predicate_rsql))
        if (len(rules) == 1 and
            ('distinct_denoted' not in rules[0]) and
            self.annotations.OkInjection(table_predicate_rsql)):
          [r] = rules
          rs = rule_translate.ExtractRuleStructure(
              r, allocator, None)
          rs.ElliminateInternalVariables(assert_full_ellimination=False)
          new_tables.update(rs.tables)
          InjectStructure(s, rs)

          new_vars_map = {}
          new_inv_vars_map = {}
          for (table_name, table_var), clause_var in s.vars_map.items():
            if table_name != table_name_rsql:
              new_vars_map[table_name, table_var] = clause_var
              new_inv_vars_map[clause_var] = (table_name, table_var)
            else:
              if table_var not in rs.select:
                if '*' in rs.select:
                  subscript = {'literal': {'the_symbol': {'symbol': table_var}}}
                  s.vars_unification.append({
                      'left': {
                          'variable': {
                              'var_name': clause_var
                          }
                      },
                      'right': {
                          'subscript': {
                              'subscript': subscript,
                              'record': rs.select['*']
                          }
                      }
                  })
                else:
                  extra_hint = '' if table_var != '*' else (
                      ' Are you using ..<rest of> for injectible predicate? '
                      'Please list the fields that you extract explicitly. '
                      'Tracking bug: b/131759583.')
                  raise rule_translate.RuleCompileException(
                      color.Format(
                          'Predicate {warning}{table_predicate_rsql}{end} '
                          'does not have an argument '
                          '{warning}{table_var}{end}, but '
                          'this rule tries to access it. {extra_hint}',
                          dict(table_predicate_rsql=table_predicate_rsql,
                               table_var=table_var,
                               extra_hint=extra_hint)),
                      s.full_rule_text)
              else:
                s.vars_unification.append({
                    'left': {'variable': {'var_name': clause_var}},
                    'right': rs.select[table_var]
                })
          s.vars_map = new_vars_map
          s.inv_vars_map = new_inv_vars_map
        else:
          new_tables[table_name_rsql] = table_predicate_rsql
      if s.tables == new_tables:
        break
      s.tables = new_tables

  def SingleRuleSql(self, rule,
                    allocator=None, external_vocabulary=None):
    """Producing SQL for a given rule in the program."""
    allocator = allocator or self.NewNamesAllocator()
    r = rule
    s = rule_translate.ExtractRuleStructure(
        r, allocator, external_vocabulary)

    s.ElliminateInternalVariables(assert_full_ellimination=False)

    self.RunInjections(s, allocator)
    s.ElliminateInternalVariables(assert_full_ellimination=True)
    s.UnificationsToConstraints()
    try:
      sql = s.AsSql(self.MakeSubqueryTranslator(allocator), self.flag_values)
    except RuntimeError as runtime_error:
      if (str(runtime_error).startswith('maximum recursion')):
        raise rule_translate.RuleCompileException(
            RecursionError(),
            s.full_rule_text)
      else:
        raise runtime_error
    if 'nil' in s.tables.values():
      # Mark rule for deletion.
      sql = '/* nil */' + sql
    return sql

  def GenerateWithClauses(self, predicate_name):
    """Generate the WITH ... prefix for queries that use it."""
    dependencies = self.execution.table_to_with_dependencies[predicate_name]

    if not dependencies:
      return None

    with_bodies = []
    for dependency in dependencies:
      table_name = self.execution.table_to_defined_table_map[dependency]
      sql = self.execution.table_to_with_sql_map[table_name]
      with_bodies.append('{} AS ({})'.format(table_name, sql))

    return 'WITH {}'.format(',\n'.join(with_bodies))

  def MakeSubqueryTranslator(self, allocator):
    return SubqueryTranslator(self, allocator, self.execution)


class SubqueryTranslator(object):
  """Converter of tables and rules in a context of a universe."""

  def __init__(self, program, allocator, execution):
    self.program = program
    self.allocator = allocator
    self.execution = execution

  def TranslateTableAttachedToFile(self, table, ground, external_vocabulary):
    """Translates file-attached table. Appends exports and defines."""
    self.execution.dependency_edges.append((
        table,
        self.execution.workflow_predicates_stack[-1]))
    if table in self.execution.table_to_defined_table_map:
      return self.execution.table_to_defined_table_map[table]
    table_name = ground.table_name
        #self.allocator.AllocateTable(hint_for_user=table)
    self.execution.table_to_defined_table_map[table] = table_name
    define_statement = '-- Interacting with table %s' % table_name
    self.execution.AddDefine(define_statement)
    export_statement = None
    if table in self.program.defined_predicates:
      self.execution.workflow_predicates_stack.append(table)
      dependency_sql = self.program.PredicateSql(
          table, self.allocator, external_vocabulary)

      # Wrap query in with
      with_signature = self.program.GenerateWithClauses(table)
      if with_signature:
        dependency_sql = '{}\n{}'.format(with_signature, dependency_sql)

      dependency_sql = self.program.UseFlagsAsParameters(dependency_sql)
      self.execution.workflow_predicates_stack.pop()
      maybe_drop_table = (
          'DROP TABLE IF EXISTS %s;\n' % ground.table_name
          if ground.overwrite else '')
      export_statement = (
          maybe_drop_table +
          'CREATE TABLE {name} AS {dependency_sql}'.format(
              name=ground.table_name,
              dependency_sql=FormatSql(dependency_sql)))

      export_statement = self.program.UseFlagsAsParameters(export_statement)
      # It's cheap to store a string multiple times in Python, as it's stored
      # via a pointer.
      self.execution.table_to_export_map[table] = export_statement
      self.execution.export_statements.append(export_statement)
    if export_statement:
      self.execution.defines_and_exports.append(export_statement)
    self.execution.defines_and_exports.append(define_statement)
    return table_name

  def TranslateWithedTable(self, table):
    """Translates table that should be defined in a WITH clause."""
    parent_table = self.execution.workflow_predicates_stack[-1]
    if table not in self.execution.table_to_defined_table_map:
      table_name = self.allocator.AllocateTable(hint_for_user=table)
      self.execution.table_to_defined_table_map[table] = table_name
      # We don't pass external vocabulary; named predicates should not have
      # free terms.
      implementation = self.program.PredicateSql(table, self.allocator)
      self.execution.table_to_with_sql_map[table_name] = implementation
    else:
      # Calling predicate SQL to add the required ground dependencies.
      if table not in self.execution.with_compilation_done_for_parent[
          parent_table]:
        # Swap these lines to compile a recursive with.
        _ = self.program.PredicateSql(table, self.allocator)
        self.execution.with_compilation_done_for_parent[
            parent_table].add(table)

    # Adding dependencies at the end means we add the deepest dependencies
    # first, which ensures our WITH clause is ordered correctly.
    # Note that even if table is already defined, we need to add a dependency
    # from parent_table; table may have been defined from a different parent
    # previously.
    if table not in self.execution.table_to_with_dependencies[parent_table]:
      self.execution.table_to_with_dependencies[parent_table].append(table)
    return self.execution.table_to_defined_table_map[table]

  @classmethod
  def UnquoteParenthesised(cls, table):
    """Enable direct usage of SQL strings as table names."""
    if len(table) > 4 and table[:2] == '`(' and table[-2:] == ')`':
      return table[2:-2]
    return table

  def TranslateTable(self, table, external_vocabulary):
    """Translating table to an SQL string in the FROM cause."""
    if table in self.program.table_aliases:
      return self.program.table_aliases[table]
    ground = self.program.annotations.Ground(table)
    if ground:
      return self.TranslateTableAttachedToFile(
          table, ground, external_vocabulary)
    if table in self.program.defined_predicates:
      if self.program.execution.With(table):
        return self.TranslateWithedTable(table)
      return '(%s)' % self.program.PredicateSql(
          table, self.allocator, external_vocabulary)
      predicate_sql = Indent2(predicate_sql)
      return '(\n%s\n)' % predicate_sql
    self.execution.data_dependency_edges.append((
      table,
      self.execution.workflow_predicates_stack[-1]))
    return self.UnquoteParenthesised(table)

  def TranslateRule(self, rule, external_vocabulary):
    return self.program.SingleRuleSql(rule, self.allocator, external_vocabulary)


def InjectStructure(target, source):
  """Injecting source RuleStructure into target."""
  target.vars_map.update(source.vars_map)
  target.inv_vars_map.update(source.inv_vars_map)
  target.vars_unification.extend(source.vars_unification)
  target.unnestings.extend(source.unnestings)
  target.constraints.extend(source.constraints)


def RecursionError():
  return color.Format(
      'The rule appears to use recursion. '
      '{warning}Recursion{end} is neither supported by '
      'Logica nor by StandardSQL.')


def RaiseCompilerError(message, context):
  raise rule_translate.RuleCompileException(message, context)


def FieldValuesAsList(field_values):
  field_values = copy.deepcopy(field_values)
  if '__rule_text' in field_values:
    del field_values['__rule_text']
  field_values_list = []
  for i in range(len(field_values)):
    i = str(i + 1)
    if i not in field_values:
      return None  # Error!
    field_values_list.append(field_values[i])
  return field_values_list




#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Library for using Logica in CoLab."""

from .common import color
from .common import concertina_lib

from .compiler import rule_translate
from .compiler import universe

import IPython

from IPython.core.magic import register_cell_magic
from IPython.display import display

import os

import pandas

from .parser_py import parse
from .common import sqlite3_logica

BQ_READY = True  # By default.

try:
  from google.cloud import bigquery
except:
  BQ_READY = False
  print('Could not import google.cloud.bigquery.')

try:
  from google.colab import auth
except:
  BQ_READY = False
  print('Could not import google.cloud.auth.')

try:
  from google.colab import widgets
  WIDGETS_IMPORTED = True
except:
  WIDGETS_IMPORTED = False
  print('Could not import google.colab.widgets.')

PROJECT = None

# TODO: Should this be renamed to PSQL_ENGINE, PSQL_CONNECTION?
DB_ENGINE = None
DB_CONNECTION = None

USER_AUTHENTICATED = False

TABULATED_OUTPUT = True

SHOW_FULL_QUERY = True

def SetProject(project):
  global PROJECT
  PROJECT = project

def SetDbConnection(connection):
  global DB_CONNECTION
  DB_CONNECTION = connection

def EnsureAuthenticatedUser():
  global USER_AUTHENTICATED
  global PROJECT
  if USER_AUTHENTICATED:
    return
  auth.authenticate_user()
  if PROJECT is None:
    print("Please enter project_id to use for BigQuery querries.")
    PROJECT = input()
    print("project_id is set to %s" % PROJECT)
    print("You can change it with logica.colab_logica.SetProject command.")
  USER_AUTHENTICATED = True

def SetTabulatedOutput(tabulated_output):
  global TABULATED_OUTPUT
  global SHOW_FULL_QUERY
  TABULATED_OUTPUT = tabulated_output
  SHOW_FULL_QUERY = TABULATED_OUTPUT

if not WIDGETS_IMPORTED:
  SetTabulatedOutput(False)

def TabBar(*args):
  """Returns a real TabBar or a mock. Useful for UIs that don't support JS."""
  if TABULATED_OUTPUT:
    return widgets.TabBar(*args)
  class MockTab:
      def __init__(self):
          pass
      def __enter__(self):
          pass
      def __exit__(self, *x):
          pass
  class MockTabBar:
      def __init__(self):
          pass
      def output_to(self, x):
          return MockTab()
  return MockTabBar()

@register_cell_magic
def logica(line, cell):
  Logica(line, cell, run_query=True)


def ParseList(line):
  line = line.strip()
  if not line:
    predicates = []
  else:
    predicates = [p.strip() for p in line.split(',')]
  return predicates


def RunSQL(sql, engine, connection=None, is_final=False):
  if engine == 'bigquery':
    EnsureAuthenticatedUser()
    client = bigquery.Client(project=PROJECT)
    return client.query(sql).to_dataframe()
  elif engine == 'psql':
    if is_final:
      return pandas.read_sql(sql, connection)
    else:
      return connection.execute(sql)
  elif engine == 'sqlite':
    statements = parse.SplitRaw(sql, ';')
    connection.executescript(sql)
    if is_final:
      return pandas.read_sql(statements[-1], connection)
    else:
      pass
    return None
  else:
    raise Exception('Logica only supports BigQuery, PostgreSQL and SQLite '
                    'for now.')


class SqliteRunner(object):
  def __init__(self):
    self.connection = sqlite3_logica.SqliteConnect()
  
  # TODO: Sqlite runner should not be accepting an engine.
  def __call__(self, sql, engine, is_final):
    return RunSQL(sql, engine, self.connection, is_final)


class PostgresRunner(object):
  def __init__(self):
    global DB_CONNECTION
    global DB_ENGINE
    if DB_CONNECTION:
      self.engine = DB_ENGINE
      self.connection = DB_CONNECTION
    else:
      (self.engine, self.connection) = PostgresJumpStart()
      DB_ENGINE = self.engine
      DB_CONNECTION = self.connection
  
  def  __call__(self, sql, engine, is_final):
    return RunSQL(sql, engine, self.connection, is_final)


def Logica(line, cell, run_query):
  """Running Logica predicates and storing results."""
  predicates = ParseList(line)
  try:
    parsed_rules = parse.ParseFile(cell)['rule']
  except parse.ParsingException as e:
    e.ShowMessage()
    return
  program = universe.LogicaProgram(parsed_rules)
  engine = program.annotations.Engine()

  if engine == 'bigquery' and not BQ_READY:
    print(color.Format(
      '[ {error}Error{end} ] BigQuery client and/or authentification is not installed. \n'
      'It is the easiest to run BigQuery requests from Google CoLab:\n'
      '  https://colab.research.google.com/.\n'
      'Note that running Logica on SQLite requires no installation.\n'
      'This could be a good fit for working with small data or learning Logica.\n'
      'Use {warning}@Engine("sqlite");{end} annotation in your program to use SQLite.'))
    return

  bar = TabBar(predicates + ['(Log)'])
  logs_idx = len(predicates)

  executions = []
  sub_bars = []
  ip = IPython.get_ipython()
  for idx, predicate in enumerate(predicates):
    with bar.output_to(logs_idx):
      try:
        sql = program.FormattedPredicateSql(predicate)
        executions.append(program.execution)
        ip.push({predicate + '_sql': sql})
      except rule_translate.RuleCompileException as e:
        print('Encountered error when compiling %s.' % predicate)
        e.ShowMessage()
        return
    # Publish output to Colab cell.
    with bar.output_to(idx):
      sub_bar = TabBar(['SQL', 'Result'])
      sub_bars.append(sub_bar)
      with sub_bar.output_to(0):
        if SHOW_FULL_QUERY:
          print(
              color.Format(
                  'The following query is stored at {warning}%s{end} '
                  'variable.' % (
                      predicate + '_sql')))
          print(sql)
        else:
          print('Query is stored at %s variable.' %
                color.Warn(predicate + '_sql'))

  with bar.output_to(logs_idx):
    if engine == 'sqlite':
      sql_runner = SqliteRunner()
    elif engine == 'psql':
      sql_runner = PostgresRunner()
    else:
      sql_runner = RunSQL
    result_map = concertina_lib.ExecuteLogicaProgram(
      executions, sql_runner=sql_runner, sql_engine=engine)

  for idx, predicate in enumerate(predicates):
    t = result_map[predicate]
    ip.push({predicate: t})
    with bar.output_to(idx):
      with sub_bars[idx].output_to(1): 
        if run_query:
          print(
              color.Format(
                  'The following table is stored at {warning}%s{end} '
                  'variable.' %
                  predicate))
          display(t)  
        else:
          print('The query was not run.')
      print(' ') # To activate the tabbar.

def PostgresJumpStart():
  # Install postgresql server.
  print("Installing and configuring an empty PostgreSQL database.")
  result = 0
  result += os.system('sudo apt-get -y -qq update')
  result += os.system('sudo apt-get -y -qq install postgresql')
  result += os.system('sudo service postgresql start')
  # Ignoring user creation error, as they may already exist.
  result += 0 * os.system(
    'sudo -u postgres psql -c "CREATE USER logica WITH SUPERUSER"')
  result += os.system(
    'sudo -u postgres psql -c "ALTER USER logica PASSWORD \'logica\';"')
  result += os.system(
    'sudo -u postgres psql -U postgres -c \'CREATE DATABASE logica;\'')
  if result != 0:
    print("""Installation failed. Please try the following manually:
# Install Logica.
!pip install logica

# Install postgresql server.
!sudo apt-get -y -qq update
!sudo apt-get -y -qq install postgresql
!sudo service postgresql start

# Prepare database for Logica.
!sudo -u postgres psql -c "CREATE USER logica WITH SUPERUSER"
!sudo -u postgres psql -c "ALTER USER logica PASSWORD 'logica';"
!sudo -u postgres psql -U postgres -c 'CREATE DATABASE logica;'

# Connect to the database.
from logica import colab_logica
from sqlalchemy import create_engine
import pandas
engine = create_engine('postgresql+psycopg2://logica:logica@127.0.0.1', pool_recycle=3600);
connection = engine.connect();
colab_logica.SetDbConnection(connection)""")
    return
  print('Installation succeeded. Connecting...')
  # Connect to the database.
  from logica import colab_logica
  from sqlalchemy import create_engine
  import pandas
  engine = create_engine('postgresql+psycopg2://logica:logica@127.0.0.1', pool_recycle=3600)
  connection = engine.connect()
  print('Connected.')
  return engine, connection


#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Utilities for YotaQL tests."""

import subprocess

if '.' not in __package__:
  from common import color
  from common import logica_lib
else:
  from ..common import color
  from ..common import logica_lib


class TestManager(object):
  """Global test managing class."""
  GOLDEN_RUN = False
  ANNOUNCE_TESTS = False
  RUN_ONLY = []

  @classmethod
  def SetGoldenRun(cls, value):
    cls.GOLDEN_RUN = value

  @classmethod
  def SetAnnounceTests(cls, value):
    cls.ANNOUNCE_TESTS = value

  @classmethod
  def SetRunOnlyTests(cls, value):
    cls.RUN_ONLY = value

  @classmethod
  def RunTest(cls, name, src, predicate, golden, user_flags,
              import_root=None):
    if cls.RUN_ONLY and name not in cls.RUN_ONLY:
      return
    RunTest(name, src, predicate, golden, user_flags,
            cls.GOLDEN_RUN, cls.ANNOUNCE_TESTS,
            import_root)


def RunTest(name, src, predicate, golden,
            user_flags=None,
            overwrite=False, announce=False,
            import_root=None):
  """Run one test."""
  if announce:
    print('Running test:', name)
  test_result = '{warning}RUNNING{end}'
  print(color.Format('% 50s   %s' % (name, test_result)))

  result = logica_lib.RunPredicate(src, predicate,
                                   user_flags=user_flags,
                                   import_root=import_root)
  # Hacky way to remove query that BQ prints.
  if '+---' in result[200:]:
    result = result[result.index('+---'):]

  if overwrite:
    with open(golden, 'w') as w:
      w.write(result)
  golden_result = open(golden).read()

  if result == golden_result:
    test_result = '{ok}PASSED{end}'
  else:
    p = subprocess.Popen(['diff', '-', golden], stdin=subprocess.PIPE)
    p.communicate(result.encode())
    test_result = '{error}FAILED{end}'

  print('\033[F\033[K' + color.Format('% 50s   %s' % (name, test_result)))


def PrintHeader():
  """Print header for all tests."""
  print(color.Format(
      '% 64s   %s' % ('{warning}TEST{end}', '{warning}RESULT{end}')))
  print(color.Format(
      '% 64s   %s' % ('{warning}----{end}', '{warning}------{end}')))




#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Built-in StandardSQL functions."""

import csv
import io

def GetCsv():
  return csv.reader(io.StringIO(CSV_DATA))

CSV_DATA = u"""function,sql_function,aggregates,has_repeated_args,min_args,max_args
$add,+,false,0,2,2
$and,AND,false,1,1,1
$array_at_offset,ARRAY[OFFSET()],false,0,2,2
$array_at_ordinal,ARRAY[ORDINAL()],false,0,2,2
$between,BETWEEN,false,0,3,3
$bitwise_and,&,false,0,2,2
$bitwise_left_shift,<<,false,0,2,2
$bitwise_not,~,false,0,1,1
$bitwise_or,|,false,0,2,2
$bitwise_right_shift,>>,false,0,2,2
$bitwise_xor,^,false,0,2,2
$case_no_value,CASE,false,2,1,1
$case_with_value,CASE,false,2,2,2
$count_star,COUNT(*),true,0,0,0
$divide,/,false,0,2,2
$equal,=,false,0,2,2
$extract,EXTRACT,false,0,2,3
$extract_date,EXTRACT,false,0,1,2
$extract_datetime,EXTRACT,false,0,1,2
$extract_time,EXTRACT,false,0,1,2
$greater,>,false,0,2,2
$greater_or_equal,>=,false,0,2,2
$in,IN,false,1,1,1
$in_array,IN UNNEST,false,0,2,2
$is_false,IS FALSE,false,0,1,1
$is_null,IS NULL,false,0,1,1
$is_true,IS TRUE,false,0,1,1
$less,<,false,0,2,2
$less_or_equal,<=,false,0,2,2
$like,LIKE,false,0,2,2
$make_array,ARRAY[...],false,1,0,0
$multiply,*,false,0,2,2
$not,NOT,false,0,1,1
$not_equal,!=,false,0,2,2
$or,OR,false,1,1,1
$safe_array_at_offset,ARRAY[SAFE_OFFSET()],false,0,2,2
$safe_array_at_ordinal,ARRAY[SAFE_ORDINAL()],false,0,2,2
$subtract,-,false,0,2,2
$unary_minus,-,false,0,1,1
abs,ABS,false,0,1,1
acos,ACOS,false,0,1,1
acosh,ACOSH,false,0,1,1
aead.decrypt_bytes,AEAD.DECRYPT_BYTES,false,0,3,3
aead.decrypt_string,AEAD.DECRYPT_STRING,false,0,3,3
aead.encrypt,AEAD.ENCRYPT,false,0,3,3
any_value,ANY_VALUE,true,0,1,1
approx_count_distinct,APPROX_COUNT_DISTINCT,true,0,1,1
approx_quantiles,APPROX_QUANTILES,true,0,2,2
approx_top_count,APPROX_TOP_COUNT,true,0,2,2
approx_top_sum,APPROX_TOP_SUM,true,0,3,3
array_agg,ARRAY_AGG,true,0,1,1
array_concat,ARRAY_CONCAT,false,1,1,1
array_concat_agg,ARRAY_CONCAT_AGG,true,0,1,1
array_length,ARRAY_LENGTH,false,0,1,1
array_reverse,ARRAY_REVERSE,false,0,1,1
array_to_string,ARRAY_TO_STRING,false,0,2,3
asin,ASIN,false,0,1,1
asinh,ASINH,false,0,1,1
atan,ATAN,false,0,1,1
atan2,ATAN2,false,0,2,2
atanh,ATANH,false,0,1,1
avg,AVG,true,0,1,1
bit_and,BIT_AND,true,0,1,1
bit_cast_to_int32,BIT_CAST_TO_INT32,false,0,1,1
bit_cast_to_int64,BIT_CAST_TO_INT64,false,0,1,1
bit_cast_to_uint32,BIT_CAST_TO_UINT32,false,0,1,1
bit_cast_to_uint64,BIT_CAST_TO_UINT64,false,0,1,1
bit_count,BIT_COUNT,false,0,1,1
bit_or,BIT_OR,true,0,1,1
bit_xor,BIT_XOR,true,0,1,1
byte_length,BYTE_LENGTH,false,0,1,1
ceil,CEIL,false,0,1,1
char_length,CHAR_LENGTH,false,0,1,1
coalesce,COALESCE,false,1,0,0
code_points_to_bytes,CODE_POINTS_TO_BYTES,false,0,1,1
code_points_to_string,CODE_POINTS_TO_STRING,false,0,1,1
concat,CONCAT,false,1,1,1
corr,CORR,true,0,2,2
cos,COS,false,0,1,1
cosh,COSH,false,0,1,1
count,COUNT,true,0,1,1
countif,COUNTIF,true,0,1,1
covar_pop,COVAR_POP,true,0,2,2
covar_samp,COVAR_SAMP,true,0,2,2
cume_dist,CUME_DIST,false,0,0,0
current_date,CURRENT_DATE,false,0,0,1
current_datetime,CURRENT_DATETIME,false,0,0,1
current_time,CURRENT_TIME,false,0,0,1
current_timestamp,CURRENT_TIMESTAMP,false,0,0,0
date,DATE,false,0,1,3
date_add,DATE_ADD,false,0,2,2
date_diff,DATE_DIFF,false,0,3,3
date_from_unix_date,DATE_FROM_UNIX_DATE,false,0,1,1
date_sub,DATE_SUB,false,0,2,2
date_trunc,DATE_TRUNC,false,0,2,2
datetime,DATETIME,false,0,1,6
datetime_add,DATETIME_ADD,false,0,2,2
datetime_diff,DATETIME_DIFF,false,0,3,3
datetime_sub,DATETIME_SUB,false,0,2,2
datetime_trunc,DATETIME_TRUNC,false,0,2,2
dense_rank,DENSE_RANK,false,0,0,0
div,DIV,false,0,2,2
ends_with,ENDS_WITH,false,0,2,2
error,ERROR,false,0,1,1
exp,EXP,false,0,1,1
farm_fingerprint,FARM_FINGERPRINT,false,0,1,1
first_value,FIRST_VALUE,false,0,1,1
floor,FLOOR,false,0,1,1
format,FORMAT,false,1,1,1
format_date,FORMAT_DATE,false,0,2,2
format_datetime,FORMAT_DATETIME,false,0,2,2
format_time,FORMAT_TIME,false,0,2,2
format_timestamp,FORMAT_TIMESTAMP,false,0,2,3
from_base32,FROM_BASE32,false,0,1,1
from_base64,FROM_BASE64,false,0,1,1
from_hex,FROM_HEX,false,0,1,1
from_proto,FROM_PROTO,false,0,1,1
generate_array,GENERATE_ARRAY,false,0,2,3
generate_date_array,GENERATE_DATE_ARRAY,false,0,2,4
generate_timestamp_array,GENERATE_TIMESTAMP_ARRAY,false,0,4,4
generate_uuid,GENERATE_UUID,false,0,0,0
greatest,GREATEST,false,1,0,0
hll_count.extract,HLL_COUNT.EXTRACT,false,0,1,1
hll_count.init,HLL_COUNT.INIT,true,0,1,2
hll_count.merge,HLL_COUNT.MERGE,true,0,1,1
hll_count.merge_partial,HLL_COUNT.MERGE_PARTIAL,true,0,1,1
ieee_divide,IEEE_DIVIDE,false,0,2,2
if,IF,false,0,3,3
ifnull,IFNULL,false,0,2,2
is_inf,IS_INF,false,0,1,1
is_nan,IS_NAN,false,0,1,1
json_extract,JSON_EXTRACT,false,0,2,2
json_extract_scalar,JSON_EXTRACT_SCALAR,false,0,2,2
json_query,JSON_QUERY,false,0,2,2
json_value,JSON_VALUE,false,0,2,2
keys.add_key_from_raw_bytes,KEYS.ADD_KEY_FROM_RAW_BYTES,false,0,3,3
keys.keyset_from_json,KEYS.KEYSET_FROM_JSON,false,0,1,1
keys.keyset_length,KEYS.KEYSET_LENGTH,false,0,1,1
keys.keyset_to_json,KEYS.KEYSET_TO_JSON,false,0,1,1
keys.new_keyset,KEYS.NEW_KEYSET,false,0,1,1
keys.rotate_keyset,KEYS.ROTATE_KEYSET,false,0,2,2
lag,LAG,false,0,1,3
last_value,LAST_VALUE,false,0,1,1
lead,LEAD,false,0,1,3
least,LEAST,false,1,0,0
length,LENGTH,false,0,1,1
ln,LN,false,0,1,1
log,LOG,false,0,1,2
log10,LOG10,false,0,1,1
logical_and,LOGICAL_AND,true,0,1,1
logical_or,LOGICAL_OR,true,0,1,1
lower,LOWER,false,0,1,1
lpad,LPAD,false,0,2,3
ltrim,LTRIM,false,0,1,2
max,MAX,true,0,1,1
md5,MD5,false,0,1,1
min,MIN,true,0,1,1
mod,MOD,false,0,2,2
net.format_ip,NET.FORMAT_IP,false,0,1,1
net.format_packed_ip,NET.FORMAT_PACKED_IP,false,0,1,1
net.host,NET.HOST,false,0,1,1
net.ip_from_string,NET.IP_FROM_STRING,false,0,1,1
net.ip_in_net,NET.IP_IN_NET,false,0,2,2
net.ip_net_mask,NET.IP_NET_MASK,false,0,2,2
net.ip_to_string,NET.IP_TO_STRING,false,0,1,1
net.ip_trunc,NET.IP_TRUNC,false,0,2,2
net.ipv4_from_int64,NET.IPV4_FROM_INT64,false,0,1,1
net.ipv4_to_int64,NET.IPV4_TO_INT64,false,0,1,1
net.make_net,NET.MAKE_NET,false,0,2,2
net.parse_ip,NET.PARSE_IP,false,0,1,1
net.parse_packed_ip,NET.PARSE_PACKED_IP,false,0,1,1
net.public_suffix,NET.PUBLIC_SUFFIX,false,0,1,1
net.reg_domain,NET.REG_DOMAIN,false,0,1,1
net.safe_ip_from_string,NET.SAFE_IP_FROM_STRING,false,0,1,1
normalize,NORMALIZE,false,0,1,2
normalize_and_casefold,NORMALIZE_AND_CASEFOLD,false,0,1,2
nth_value,NTH_VALUE,false,0,2,2
ntile,NTILE,false,0,1,1
nullif,NULLIF,false,0,2,2
parse_date,PARSE_DATE,false,0,2,2
parse_datetime,PARSE_DATETIME,false,0,2,2
parse_time,PARSE_TIME,false,0,2,2
parse_timestamp,PARSE_TIMESTAMP,false,0,2,3
percent_rank,PERCENT_RANK,false,0,0,0
percentile_cont,PERCENTILE_CONT,false,0,2,2
percentile_disc,PERCENTILE_DISC,false,0,2,2
pow,POW,false,0,2,2
proto_default_if_null,PROTO_DEFAULT_IF_NULL,false,0,1,1
rand,RAND,false,0,0,0
range_bucket,RANGE_BUCKET,false,0,2,2
rank,RANK,false,0,0,0
regexp_contains,REGEXP_CONTAINS,false,0,2,2
regexp_extract,REGEXP_EXTRACT,false,0,2,2
regexp_extract_all,REGEXP_EXTRACT_ALL,false,0,2,2
regexp_match,REGEXP_MATCH,false,0,2,2
regexp_replace,REGEXP_REPLACE,false,0,3,3
repeat,REPEAT,false,0,2,2
replace,REPLACE,false,0,3,3
reverse,REVERSE,false,0,1,1
round,ROUND,false,0,1,2
row_number,ROW_NUMBER,false,0,0,0
rpad,RPAD,false,0,2,3
rtrim,RTRIM,false,0,1,2
safe_add,SAFE_ADD,false,0,2,2
safe_convert_bytes_to_string,SAFE_CONVERT_BYTES_TO_STRING,false,0,1,1
safe_divide,SAFE_DIVIDE,false,0,2,2
safe_multiply,SAFE_MULTIPLY,false,0,2,2
safe_negate,SAFE_NEGATE,false,0,1,1
safe_subtract,SAFE_SUBTRACT,false,0,2,2
session_user,SESSION_USER,false,0,0,0
sha1,SHA1,false,0,1,1
sha256,SHA256,false,0,1,1
sha512,SHA512,false,0,1,1
sign,SIGN,false,0,1,1
sin,SIN,false,0,1,1
sinh,SINH,false,0,1,1
split,SPLIT,false,0,1,2
sqrt,SQRT,false,0,1,1
st_accum,ST_ACCUM,true,0,1,1
st_area,ST_AREA,false,0,1,2
st_asbinary,ST_ASBINARY,false,0,1,1
st_asgeojson,ST_ASGEOJSON,false,0,1,2
st_askml,ST_ASKML,false,0,1,1
st_astext,ST_ASTEXT,false,0,1,1
st_boundary,ST_BOUNDARY,false,0,1,1
st_buffer,ST_BUFFER,false,0,2,4
st_bufferwithtolerance,ST_BUFFERWITHTOLERANCE,false,0,3,4
st_centroid,ST_CENTROID,false,0,1,1
st_centroid_agg,ST_CENTROID_AGG,true,0,1,1
st_closestpoint,ST_CLOSESTPOINT,false,0,2,3
st_contains,ST_CONTAINS,false,0,2,2
st_coveredby,ST_COVEREDBY,false,0,2,2
st_covers,ST_COVERS,false,0,2,2
st_difference,ST_DIFFERENCE,false,0,2,2
st_dimension,ST_DIMENSION,false,0,1,1
st_disjoint,ST_DISJOINT,false,0,2,2
st_distance,ST_DISTANCE,false,0,2,3
st_dwithin,ST_DWITHIN,false,0,3,4
st_equals,ST_EQUALS,false,0,2,2
st_geogfromgeojson,ST_GEOGFROMGEOJSON,false,0,1,1
st_geogfromkml,ST_GEOGFROMKML,false,0,1,1
st_geogfromtext,ST_GEOGFROMTEXT,false,0,1,2
st_geogfromwkb,ST_GEOGFROMWKB,false,0,1,1
st_geogpoint,ST_GEOGPOINT,false,0,2,2
st_geogpointfromgeohash,ST_GEOGPOINTFROMGEOHASH,false,0,1,1
st_geohash,ST_GEOHASH,false,0,1,2
st_intersection,ST_INTERSECTION,false,0,2,2
st_intersects,ST_INTERSECTS,false,0,2,2
st_intersectsbox,ST_INTERSECTSBOX,false,0,5,5
st_iscollection,ST_ISCOLLECTION,false,0,1,1
st_isempty,ST_ISEMPTY,false,0,1,1
st_length,ST_LENGTH,false,0,1,2
st_makeline,ST_MAKELINE,false,0,1,2
st_makepolygon,ST_MAKEPOLYGON,false,0,1,2
st_makepolygonoriented,ST_MAKEPOLYGONORIENTED,false,0,1,1
st_maxdistance,ST_MAXDISTANCE,false,0,2,3
st_numpoints,ST_NUMPOINTS,false,0,1,1
st_perimeter,ST_PERIMETER,false,0,1,2
st_simplify,ST_SIMPLIFY,false,0,2,2
st_snaptogrid,ST_SNAPTOGRID,false,0,2,2
st_touches,ST_TOUCHES,false,0,2,2
st_unaryunion,ST_UNARYUNION,false,0,1,1
st_union,ST_UNION,false,0,1,2
st_union_agg,ST_UNION_AGG,true,0,1,1
st_within,ST_WITHIN,false,0,2,2
st_x,ST_X,false,0,1,1
st_y,ST_Y,false,0,1,1
starts_with,STARTS_WITH,false,0,2,2
stddev_pop,STDDEV_POP,true,0,1,1
stddev_samp,STDDEV_SAMP,true,0,1,1
string,STRING,false,0,1,2
string_agg,STRING_AGG,true,0,1,2
strpos,STRPOS,false,0,2,2
substr,SUBSTR,false,0,2,3
sum,SUM,true,0,1,1
tan,TAN,false,0,1,1
tanh,TANH,false,0,1,1
time,TIME,false,0,1,3
time_add,TIME_ADD,false,0,2,2
time_diff,TIME_DIFF,false,0,3,3
time_sub,TIME_SUB,false,0,2,2
time_trunc,TIME_TRUNC,false,0,2,2
timestamp,TIMESTAMP,false,0,1,2
timestamp_add,TIMESTAMP_ADD,false,0,2,2
timestamp_diff,TIMESTAMP_DIFF,false,0,3,3
timestamp_from_unix_micros,TIMESTAMP_FROM_UNIX_MICROS,false,0,1,1
timestamp_from_unix_millis,TIMESTAMP_FROM_UNIX_MILLIS,false,0,1,1
timestamp_from_unix_seconds,TIMESTAMP_FROM_UNIX_SECONDS,false,0,1,1
timestamp_micros,TIMESTAMP_MICROS,false,0,1,1
timestamp_millis,TIMESTAMP_MILLIS,false,0,1,1
timestamp_seconds,TIMESTAMP_SECONDS,false,0,1,1
timestamp_sub,TIMESTAMP_SUB,false,0,2,2
timestamp_trunc,TIMESTAMP_TRUNC,false,0,2,3
to_base32,TO_BASE32,false,0,1,1
to_base64,TO_BASE64,false,0,1,1
to_code_points,TO_CODE_POINTS,false,0,1,1
to_hex,TO_HEX,false,0,1,1
to_json_string,TO_JSON_STRING,false,0,1,2
to_proto,TO_PROTO,false,0,1,1
trim,TRIM,false,0,1,2
trunc,TRUNC,false,0,1,2
unix_date,UNIX_DATE,false,0,1,1
unix_micros,UNIX_MICROS,false,0,1,1
unix_millis,UNIX_MILLIS,false,0,1,1
unix_seconds,UNIX_SECONDS,false,0,1,1
upper,UPPER,false,0,1,1
var_pop,VAR_POP,true,0,1,1
var_samp,VAR_SAMP,true,0,1,1
""";



#!/usr/bin/python
#
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Basic VisJS graph visualization wrapper."""

from IPython.display import display, HTML
import urllib
import json
import numpy

visjs = urllib.request.urlopen('https://raw.githubusercontent.com/EvgSkv/vis/master/dist/vis-network.min.js').read()
css = urllib.request.urlopen('https://raw.githubusercontent.com/EvgSkv/vis/master/dist/vis-network.min.css').read()

RENDERED_GRAPHS = 0

def GraphHtml(nodes, edges, options, width, height):
  html_colab = r"""
  <script type="text/javascript">
  %(visjs)s
  </script>
  <style>
  %(css)s
  </style>
  <style type="text/css">
  #graph_%(idx)s {
    width: %(width)spx;
    height: %(height)spx;
    border: 1px solid lightgray;
  }
  </style>
  <div id="graph_%(idx)s"></div>

  <script type="text/javascript">
    // Nodes array.
    var nodes = new vis.DataSet(%(nodes)s);

    // Edges array.
    var edges = new vis.DataSet(%(edges)s);

    // Network.
    var container = document.getElementById('graph_%(idx)s');
    var data = {
      nodes: nodes,
      edges: edges
    };
    var options = %(options)s;
    var network = new vis.Network(container, data, options);
  </script>
  """

  html_jupyter = r"""
  <style>
  %(css)s
  </style>
  <style type="text/css">
  #graph_%(idx)s {
    width: %(width)spx;
    height: %(height)spx;
    border: 1px solid lightgray;
  }
  </style>
  <div id="graph_%(idx)s"></div>

  <script type="text/javascript">
    require(['https://evgskv.github.io/vis/dist/vis.js'], function(vis) {
      // Nodes array.
      var nodes = new vis.DataSet(%(nodes)s);

      // Edges array.
      var edges = new vis.DataSet(%(edges)s);

      // Network.
      var container = document.getElementById('graph_%(idx)s');
      var data = {
        nodes: nodes,
        edges: edges
      };
      var options = %(options)s;
      var network = new vis.Network(container, data, options);
    });
  </script>
  """
  if 'google.colab' in str(get_ipython()):
    html = html_colab
  else:
    html = html_jupyter
  global RENDERED_GRAPHS
  idx = RENDERED_GRAPHS
  RENDERED_GRAPHS = RENDERED_GRAPHS + 1
  result = html % dict(visjs=visjs.decode(), css=css.decode(),
                       nodes=json.dumps(nodes),
                       edges=json.dumps(edges),
                       options=json.dumps(options),
                       height=height,
                       width=width,
                       idx=idx)
  return result

def DisplayGraph(nodes, edges, options=None, width=640, height=480):
  options = options or {}
  html = GraphHtml(nodes, edges, options, width, height)
  display(HTML(html))

def SimpleGraphFromList(edges):
  nodes = list({n for e in edges for n in e})
  nodes_json = [{"id": n, "label": str(n)} for n in nodes]
  edges_json = [{"from": e[0], "to": e[1]} for e in edges]
  DisplayGraph(nodes_json, edges_json)

def SimpleGraph(p, source='col0', target='col1'):
  edges = list(zip(p[source], p[target]))
  SimpleGraphFromList(edges)

def DirectedGraphFromList(edges):
  nodes = list({n for e in edges for n in e})
  nodes_json = [{"id": n, "label": str(n)} for n in nodes]
  edges_json = [{"from": e[0], "to": e[1], "arrows": "to"}
                for e in edges]
  DisplayGraph(nodes_json, edges_json)

def DirectedGraph(p, source='col0', target='col1'):
  edges = list(zip(p[source], p[target]))
  DirectedGraphFromList(edges)

def Graph(nodes, edges, options=None, width=640, height=480):
  nodes_list = [dict(n) for _, n in nodes.iterrows()]
  edges_list = [dict(e) for _, e in edges.iterrows()]
  def Convert(l):
    for e in l:
      for k in list(e):
        if isinstance(e[k], numpy.int64):
          e[k] = int(e[k])
        if k == 'source':
          e['from'] = e[k]
        if k == 'target':
          e['to'] = e[k]
        if k == 'class':
          e['group'] = e[k]
  Convert(nodes_list)
  Convert(edges_list)
  options_obj = {} if options is None else dict(options.iloc[0])
  Convert([options_obj])
  DisplayGraph(nodes_list, edges_list, options_obj, width, height)
